import json

from sirius_sdk.encryption import P2PConnection
from sirius_sdk.base import ReadOnlyChannel, WriteOnlyChannel
from sirius_sdk.messaging import Message
from sirius_sdk.errors.exceptions import *


class AddressedTunnel:
    """Transport abstraction that help build tunnels (p2p pairwise relationships) over channel layer.
    """

    ENC = 'utf-8'

    class Context:
        """Tunnel instance context"""
        def __init__(self):
            # encrypted: flag that represent that last received message was encrypted
            self.encrypted = False

    def __init__(self, address: str, input_: ReadOnlyChannel, output_: WriteOnlyChannel, p2p: P2PConnection):
        """
        :param address: communication address of transport environment on server-side
        :param input_: channel of input stream
        :param output_: channel of output stream
        :param p2p: pairwise connection that configured and prepared outside
        """
        self.__address = address
        self.__input = input_
        self.__output = output_
        self.__p2p = p2p
        self.__context = self.Context()

    @property
    def address(self):
        return self.__address

    @property
    def context(self):
        return self.__context

    async def receive(self, timeout: int=None) -> Message:
        """
        Read message.

        Tunnel allows to receive non-encrypted messages, high-level logic may control message encryption flag
        via context.encrypted field

        :param timeout:timeout in seconds
        :return: received packet
        """
        payload = await self.__input.read(timeout)
        if not isinstance(payload, bytes) and not isinstance(payload, dict):
            raise TypeError('Expected bytes or dict, got {}'.format(type(payload)))
        if isinstance(payload, bytes):
            try:
                payload = json.loads(payload)
            except Exception as e:
                raise SiriusInvalidPayloadStructure("Invalid packed message") from e
        if 'protected' in payload:
            unpacked = self.__p2p.unpack(payload)
            self.__context.encrypted = True
            return Message(unpacked)
        else:
            self.__context.encrypted = False
            return Message(payload)

    async def post(self, message: Message, encrypt: bool=True) -> bool:
        """Write message

        :param message: message to send
        :param encrypt: do encryption
        :return: operation success
        """
        if encrypt:
            payload = self.__p2p.pack(message)
        else:
            payload = message.serialize().encode(self.ENC)
        return await self.__output.write(payload)



import uuid

import pytest

from sirius_sdk import Agent
from sirius_sdk.messaging import Message, register_message_class
from .helpers import ServerTestSuite


class TrustPingMessageUnderTest(Message):
    pass


@pytest.mark.asyncio
async def test__all_agents_ping(test_suite: ServerTestSuite):
    for name in ['agent1', 'agent2', 'agent3', 'agent4']:
        params = test_suite.get_agent_params(name)
        agent = Agent(
            server_address=params['server_address'],
            credentials=params['credentials'],
            p2p=params['p2p'],
            timeout=5,
        )
        await agent.open()
        try:
            success = await agent.ping()
            assert success is True, 'agent [%s] is not ping-able' % name
        finally:
            await agent.close()


@pytest.mark.asyncio
async def test_agents_wallet(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    agent = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    await agent.open()
    try:
        # Check wallet calls is ok
        did, verkey = await agent.wallet.did.create_and_store_my_did()
        assert did
        assert verkey
        # check reopen is OK
        await agent.reopen()
    finally:
        await agent.close()


@pytest.mark.asyncio
async def test_agents_communications(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send_message(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping'
        assert msg['@id'] == trust_ping.id
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_listener_restore_message(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Bind Message class to protocol
        register_message_class(TrustPingMessageUnderTest, protocol='trust_ping_test')
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping_test/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send_message(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        msg = event['message']
        assert isinstance(msg, TrustPingMessageUnderTest), 'Unexpected msg type: ' + str(type(msg))
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_agents_trust_ping(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent1_listener = await agent1.subscribe()
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send_message(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        assert event['recipient_verkey'] == entity2['verkey']
        assert event['sender_verkey'] == entity1['verkey']
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping'
        assert msg['@id'] == trust_ping.id

        ping_response = Message(
            {
                "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping_response",
                "@id": "e002518b-456e-b3d5-de8e-7a86fe472847",
                "~thread": {"thid": trust_ping.id},
                "comment": "Hi yourself. I'm here."
            }
        )
        await agent2.send_message(
            message=ping_response,
            their_vk=entity1['verkey'],
            endpoint=agent1_endpoint,
            my_vk=entity2['verkey'],
            routing_keys=[]
        )

        event = await agent1_listener.get_one(timeout=5)
        assert event['recipient_verkey'] == entity1['verkey']
        assert event['sender_verkey'] == entity2['verkey']
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping_response'
        assert msg['@id'] == ping_response.id

    finally:
        await agent1.close()
        await agent2.close()



import asyncio
import threading
from typing import Optional
from contextlib import asynccontextmanager

from sirius_sdk.encryption.p2p import P2PConnection
from sirius_sdk.errors.exceptions import SiriusInitializationError
from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.anoncreds import AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.storages import AbstractImmutableCollection
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList
from sirius_sdk.agent.agent import Agent, BaseAgentConnection, SpawnStrategy
from .context import get as context_get, set as context_set, clear as context_clear


__ROOT_HUB = None
__THREAD_LOCAL_HUB = threading.local()


class Hub:

    def __init__(
            self, server_uri: str, credentials: bytes, p2p: P2PConnection, io_timeout: int = None,
            storage: AbstractImmutableCollection = None, crypto: AbstractCrypto = None,
            microledgers: AbstractMicroledgerList = None, pairwise_storage: AbstractPairwiseList = None,
            did: AbstractDID = None, anoncreds: AbstractAnonCreds = None, non_secrets: AbstractNonSecrets = None,
            loop: asyncio.AbstractEventLoop = None
    ):
        self.__crypto = crypto
        self.__microledgers = microledgers
        self.__pairwise_storage = pairwise_storage
        self.__did = did
        self.__anoncreds = anoncreds
        self.__non_secrets = non_secrets
        self.__server_uri = server_uri
        self.__credentials = credentials
        self.__p2p = p2p
        self.__timeout = io_timeout or BaseAgentConnection.IO_TIMEOUT
        self.__storage = storage
        self.__loop = loop or asyncio.get_event_loop()
        self.__create_agent_instance()

    def __del__(self):
        if self.__agent.is_open and self.__loop.is_running():
            asyncio.ensure_future(self.__agent.close(), loop=self.__loop)

    def copy(self):
        inst = Hub(
            server_uri=self.__server_uri, credentials=self.__credentials, p2p=self.__p2p
        )
        inst.__crypto = self.__crypto
        inst.__microledgers = self.__microledgers
        inst.__pairwise_storage = self.__pairwise_storage
        inst.__did = self.__did
        return inst

    async def abort(self):
        if self.__loop.is_running():
            if self.__loop == asyncio.get_event_loop():
                old_agent = self.__agent
                self.__create_agent_instance()
                if old_agent.is_open:
                    await old_agent.close()
            else:
                asyncio.ensure_future(self.abort(), loop=self.__loop)

    def run_soon(self, coro):
        assert asyncio.iscoroutine(coro), 'Expected coroutine object'
        if self.__loop.is_running():
            asyncio.ensure_future(coro, loop=self.__loop)

    @asynccontextmanager
    async def get_agent_connection_lazy(self):
        if not self.__agent.is_open:
            await self.__agent.open()
        yield self.__agent

    async def open(self):
        async with self.get_agent_connection_lazy() as agent:
            pass

    async def close(self):
        if self.__agent.is_open:
            await self.__agent.close()

    async def get_crypto(self) -> AbstractCrypto:
        async with self.get_agent_connection_lazy() as agent:
            return self.__crypto or agent.wallet.crypto

    async def get_microledgers(self) -> AbstractMicroledgerList:
        async with self.get_agent_connection_lazy() as agent:
            return self.__microledgers or agent.microledgers

    async def get_pairwise_list(self) -> AbstractPairwiseList:
        async with self.get_agent_connection_lazy() as agent:
            return self.__pairwise_storage or agent.pairwise_list

    async def get_did(self) -> AbstractDID:
        async with self.get_agent_connection_lazy() as agent:
            return self.__did or agent.wallet.did

    async def get_anoncreds(self) -> AbstractAnonCreds:
        async with self.get_agent_connection_lazy() as agent:
            return self.__anoncreds or agent.wallet.anoncreds

    async def get_cache(self) -> AbstractCache:
        async with self.get_agent_connection_lazy() as agent:
            return self.__anoncreds or agent.wallet.cache

    async def get_non_secrets(self) -> AbstractNonSecrets:
        async with self.get_agent_connection_lazy() as agent:
            return self.__non_secrets or agent.wallet.non_secrets

    def __create_agent_instance(self):
        self.__agent = Agent(
            server_address=self.__server_uri,
            credentials=self.__credentials,
            p2p=self.__p2p,
            timeout=self.__timeout,
            loop=self.__loop,
            storage=self.__storage,
            spawn_strategy=SpawnStrategy.CONCURRENT
        )


def init(server_uri: str, credentials: bytes, p2p: P2PConnection, io_timeout: int = None,
         storage: AbstractImmutableCollection = None,
         crypto: AbstractCrypto = None, microledgers: AbstractMicroledgerList = None,
         did: AbstractDID = None, pairwise_storage: AbstractPairwiseList = None, non_secrets: AbstractNonSecrets = None
         ):
    global __ROOT_HUB
    root = Hub(
        server_uri=server_uri, credentials=credentials, p2p=p2p, io_timeout=io_timeout,
        storage=storage, crypto=crypto, microledgers=microledgers,
        pairwise_storage=pairwise_storage, did=did, non_secrets=non_secrets
    )
    loop = asyncio.get_event_loop()
    if loop.is_running():
        raise SiriusInitializationError('You must call this method outside coroutine')
    loop.run_until_complete(root.open())
    __ROOT_HUB = root


@asynccontextmanager
async def context(
          server_uri: str, credentials: bytes, p2p: P2PConnection, io_timeout: int = None,
          storage: AbstractImmutableCollection = None,
          crypto: AbstractCrypto = None, microledgers: AbstractMicroledgerList = None,
          did: AbstractDID = None, pairwise_storage: AbstractPairwiseList = None,
          non_secrets: AbstractNonSecrets = None
):
    hub = Hub(
        server_uri=server_uri, credentials=credentials, p2p=p2p, io_timeout=io_timeout,
        storage=storage, crypto=crypto, microledgers=microledgers,
        pairwise_storage=pairwise_storage, did=did, non_secrets=non_secrets
    )
    old_hub = __get_thread_local_gub()
    __THREAD_LOCAL_HUB.instance = hub
    try:
        await hub.open()
        old_hub_coro = context_get('hub')
        context_set('hub', hub)
        try:
            yield
        finally:
            context_clear()
            await hub.close()
            context_set('hub', old_hub_coro)
    finally:
        __THREAD_LOCAL_HUB.instance = old_hub


def __get_root_hub() -> Optional[Hub]:
    return __ROOT_HUB


def __get_thread_local_gub() -> Optional[Hub]:
    # For now thread local context not used but code prepared for future purposes
    try:
        inst = __THREAD_LOCAL_HUB.instance
    except AttributeError:
        inst = None
    return inst


def _current_hub() -> Hub:
    inst = context_get('hub')
    if inst is None:
        root_hub = __get_thread_local_gub() or __get_root_hub()
        if root_hub is None:
            raise SiriusInitializationError('Non initialized Sirius Agent connection')
        inst = root_hub.copy()
        context_set('hub', inst)
    return inst



import uuid
import asyncio

import pytest

import sirius_sdk
from sirius_sdk import Agent
from sirius_sdk.agent.coprotocols import *
from .conftest import get_pairwise
from .helpers import run_coroutines
from .helpers import ServerTestSuite


TEST_MSG_TYPES = [
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/request-1',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/response-1',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/request-2',
    'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test_protocol/1.0/response-2',
]
MSG_LOG = []


async def routine1(protocol: AbstractCoProtocolTransport):
    first_req = Message({
        '@type': TEST_MSG_TYPES[0],
        'content': 'Request1'
    })
    MSG_LOG.append(first_req)
    ok, resp1 = await protocol.switch(message=first_req)
    assert ok is True
    MSG_LOG.append(resp1)
    ok, resp2 = await protocol.switch(
        message=Message({
            '@type': TEST_MSG_TYPES[2],
            'content': 'Request2'
        })
    )
    assert ok is True
    MSG_LOG.append(resp2)


async def routine1_on_hub(
        co: sirius_sdk.AbstractP2PCoProtocol,
        server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        **kwargs
):
    async with sirius_sdk.context(server_uri=server_address, credentials=credentials, p2p=p2p):
        first_req = Message({
            '@type': TEST_MSG_TYPES[0],
            'content': 'Request1'
        })
        MSG_LOG.append(first_req)
        ok, resp1 = await co.switch(message=first_req)
        assert ok is True
        MSG_LOG.append(resp1)
        ok, resp2 = await co.switch(
            message=Message({
                '@type': TEST_MSG_TYPES[2],
                'content': 'Request2'
            })
        )
        assert ok is True
        MSG_LOG.append(resp2)


async def routine2(protocol: AbstractCoProtocolTransport):
    await asyncio.sleep(1)
    ok, resp1 = await protocol.switch(
        message=Message({
            '@type': TEST_MSG_TYPES[1],
            'content': 'Response1'
        })
    )
    assert ok is True
    MSG_LOG.append(resp1)
    await protocol.send(
        message=Message({
            '@type': TEST_MSG_TYPES[3],
            'content': 'End'
        })
    )


async def routine2_on_hub(
        co: sirius_sdk.AbstractP2PCoProtocol,
        server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection,
        **kwargs
):
    async with sirius_sdk.context(server_uri=server_address, credentials=credentials, p2p=p2p):
        await asyncio.sleep(1)
        ok, resp1 = await co.switch(
            message=Message({
                '@type': TEST_MSG_TYPES[1],
                'content': 'Response1'
            })
        )
        assert ok is True
        MSG_LOG.append(resp1)
        await co.send(
            message=Message({
                '@type': TEST_MSG_TYPES[3],
                'content': 'End'
            })
        )


def check_msg_log():
    assert len(MSG_LOG) == len(TEST_MSG_TYPES)
    for i, item in enumerate(TEST_MSG_TYPES):
        assert MSG_LOG[i].type == TEST_MSG_TYPES[i]
    assert MSG_LOG[0]['content'] == 'Request1'
    assert MSG_LOG[1]['content'] == 'Response1'
    assert MSG_LOG[2]['content'] == 'Request2'
    assert MSG_LOG[3]['content'] == 'End'


@pytest.mark.asyncio
async def test__their_endpoint_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Make protocol instances
        their1 = TheirEndpoint(agent2_endpoint, entity2['verkey'])
        agent1_protocol = await agent1.spawn(entity1['verkey'], their1)
        assert isinstance(agent1_protocol, TheirEndpointCoProtocolTransport)
        their2 = TheirEndpoint(agent1_endpoint, entity1['verkey'])
        agent2_protocol = await agent2.spawn(entity2['verkey'], their2)
        assert isinstance(agent2_protocol, TheirEndpointCoProtocolTransport)
        await agent1_protocol.start(['test_protocol'])
        await agent2_protocol.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__their_endpoint_protocol_on_hub(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]

    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
    finally:
        await agent1.close()
        await agent2.close()
    # FIRE!!!
    their1 = TheirEndpoint(agent2_endpoint, entity2['verkey'])
    their2 = TheirEndpoint(agent1_endpoint, entity1['verkey'])
    co1 = sirius_sdk.CoProtocolP2PAnon(entity1['verkey'], their1, ['test_protocol'])
    co2 = sirius_sdk.CoProtocolP2PAnon(entity2['verkey'], their2, ['test_protocol'])
    MSG_LOG.clear()
    await run_coroutines(
        routine1_on_hub(co1, **agent1_params),
        routine2_on_hub(co2, **agent2_params)
    )
    check_msg_log()


@pytest.mark.asyncio
async def test__pairwise_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        agent1_protocol = await agent1.spawn(pairwise1)
        agent2_protocol = await agent2.spawn(pairwise2)
        assert isinstance(agent1_protocol, PairwiseCoProtocolTransport)
        assert isinstance(agent2_protocol, PairwiseCoProtocolTransport)

        await agent1_protocol.start(['test_protocol'])
        await agent2_protocol.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__pairwise_protocol_on_hub(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )
    finally:
        await agent1.close()
        await agent2.close()

    co1 = sirius_sdk.CoProtocolP2P(pairwise1, ['test_protocol'])
    co2 = sirius_sdk.CoProtocolP2P(pairwise2, ['test_protocol'])
    MSG_LOG.clear()
    await run_coroutines(
        routine1_on_hub(co1, **agent1_params),
        routine2_on_hub(co2, **agent2_params)
    )
    check_msg_log()


@pytest.mark.asyncio
async def test__threadbased_protocol(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        thread_id = uuid.uuid4().hex
        agent1_protocol = await agent1.spawn(thread_id, pairwise1)
        agent2_protocol = await agent2.spawn(thread_id, pairwise2)
        assert isinstance(agent1_protocol, ThreadBasedCoProtocolTransport)
        assert isinstance(agent2_protocol, ThreadBasedCoProtocolTransport)

        await agent1_protocol.start()
        await agent2_protocol.start()
        try:
            MSG_LOG.clear()
            await run_coroutines(routine1(agent1_protocol), routine2(agent2_protocol))
            check_msg_log()
        finally:
            await agent1_protocol.stop()
            await agent2_protocol.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test__threadbased_protocol_on_hub(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )
    finally:
        await agent1.close()
        await agent2.close()

    thread_id = uuid.uuid4().hex
    co1 = sirius_sdk.CoProtocolThreadedP2P(thread_id, pairwise1)
    co2 = sirius_sdk.CoProtocolThreadedP2P(thread_id, pairwise2)
    MSG_LOG.clear()
    await run_coroutines(
        routine1_on_hub(co1, **agent1_params),
        routine2_on_hub(co2, **agent2_params)
    )
    check_msg_log()


@pytest.mark.asyncio
async def test__protocols_intersections(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=15,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=15,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        # Init pairwise list #1
        did1, verkey1 = await agent1.wallet.did.create_and_store_my_did()
        did2, verkey2 = await agent2.wallet.did.create_and_store_my_did()
        await agent1.wallet.did.store_their_did(did2, verkey2)
        await agent1.wallet.pairwise.create_pairwise(
            their_did=did2, my_did=did1
        )
        await agent2.wallet.did.store_their_did(did1, verkey1)
        await agent2.wallet.pairwise.create_pairwise(
            their_did=did1, my_did=did2
        )
        # Init pairwise list #2
        pairwise1 = Pairwise(
            me=Pairwise.Me(
                did=did1,
                verkey=verkey1
            ),
            their=Pairwise.Their(
                did=did2,
                label='Label-2',
                endpoint=agent2_endpoint,
                verkey=verkey2
            )
        )
        pairwise2 = Pairwise(
            me=Pairwise.Me(
                did=did2,
                verkey=verkey2
            ),
            their=Pairwise.Their(
                did=did1,
                label='Label-1',
                endpoint=agent1_endpoint,
                verkey=verkey1
            )
        )

        thread_id = uuid.uuid4().hex
        agent1_protocol_threaded = await agent1.spawn(thread_id, pairwise1)
        agent2_protocol_threaded = await agent2.spawn(thread_id, pairwise2)
        assert isinstance(agent1_protocol_threaded, ThreadBasedCoProtocolTransport)
        assert isinstance(agent2_protocol_threaded, ThreadBasedCoProtocolTransport)
        agent1_protocol_pairwise = await agent1.spawn(pairwise1)
        agent2_protocol_pairwise = await agent2.spawn(pairwise2)
        assert isinstance(agent1_protocol_pairwise, PairwiseCoProtocolTransport)
        assert isinstance(agent2_protocol_pairwise, PairwiseCoProtocolTransport)

        await agent1_protocol_threaded.start(['test_protocol'])
        await agent2_protocol_threaded.start(['test_protocol'])
        await agent1_protocol_pairwise.start(['test_protocol'])
        await agent2_protocol_pairwise.start(['test_protocol'])
        try:
            MSG_LOG.clear()
            await run_coroutines(
                routine1(agent1_protocol_threaded), routine2(agent2_protocol_threaded),
                routine1(agent1_protocol_pairwise), routine2(agent2_protocol_pairwise),
            )
            # collect messages
            threaded_sequence = [msg for msg in MSG_LOG if '~thread' in msg]
            non_threaded_sequence = [msg for msg in MSG_LOG if '~thread' not in msg]
            # threaded messages
            MSG_LOG.clear()
            MSG_LOG.extend(threaded_sequence)
            check_msg_log()
            for msg in threaded_sequence:
                assert msg['~thread']['thid'] == thread_id
            # non-threaded messages
            MSG_LOG.clear()
            MSG_LOG.extend(non_threaded_sequence)
            check_msg_log()
        finally:
            await agent1_protocol_threaded.stop()
            await agent2_protocol_threaded.stop()
            await agent1_protocol_pairwise.stop()
            await agent2_protocol_pairwise.stop()
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_coprotocol_abort(test_suite: ServerTestSuite, agent1: Agent, agent2: Agent):
    agent1_params = test_suite.get_agent_params('agent1')
    await agent1.open()
    await agent2.open()
    try:
        pw1 = await get_pairwise(agent1, agent2)
    finally:
        await agent1.close()
        await agent2.close()

    co = sirius_sdk.CoProtocolThreadedP2P(thid=uuid.uuid4().hex, to=pw1)
    exc = None

    async def infinite_reader(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal co
        nonlocal exc
        try:
            async with sirius_sdk.context(server_address, credentials, p2p):
                while True:
                    msg = await co.get_one()
                    print(str(msg))
        except OperationAbortedManually as e:
            exc = e

    async def delayed_aborter():
        nonlocal co
        await asyncio.sleep(3)
        await co.abort()

    await run_coroutines(
        infinite_reader(**agent1_params),
        delayed_aborter(),
    )
    assert exc is not None
    assert isinstance(exc, OperationAbortedManually)


@pytest.mark.asyncio
async def test_coprotocol_abort_multiple_ops_single_hub(test_suite: ServerTestSuite, agent1: Agent, agent2: Agent):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    await agent1.open()
    await agent2.open()
    try:
        pw1 = await get_pairwise(agent1, agent2)
        pw2 = await get_pairwise(agent2, agent1)
    finally:
        await agent1.close()
        await agent2.close()

    co = sirius_sdk.CoProtocolThreadedP2P(thid=uuid.uuid4().hex, to=pw1)
    new_thread_id = 'new-thread-id-' + uuid.uuid4().hex
    msg_log = []

    async def coroutine1(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal co
        nonlocal pw1
        nonlocal msg_log
        nonlocal new_thread_id
        async with sirius_sdk.context(server_address, credentials, p2p):
            try:
                while True:
                    msg = await co.get_one()
                    print(str(msg))
            except OperationAbortedManually:
                pass

            try:
                new_co_on_same_hub = sirius_sdk.CoProtocolThreadedP2P(thid=new_thread_id, to=pw1)
                msg, _, _ = await new_co_on_same_hub.get_one()
                print('!')
                msg_log.append(msg)
            except Exception as e:
                raise

    async def coroutine2(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal co
        nonlocal pw1
        nonlocal new_thread_id
        await asyncio.sleep(3)

        await co.abort()

        await asyncio.sleep(3)
        async with sirius_sdk.context(server_address, credentials, p2p):
            try:
                new_co_on_same_hub = sirius_sdk.CoProtocolThreadedP2P(thid=new_thread_id, to=pw2)
                await new_co_on_same_hub.send(
                    sirius_sdk.aries_rfc.Ping(comment='Test Ping')
                )
            except Exception as e:
                raise

    await run_coroutines(
        coroutine1(**agent1_params),
        coroutine2(**agent2_params)
    )
    assert len(msg_log) == 1
    msg = msg_log[0]
    assert isinstance(msg, sirius_sdk.aries_rfc.Ping)
    assert msg.comment == 'Test Ping'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_send(
        test_suite: ServerTestSuite, agent1: Agent, agent2: Agent, agent3: Agent
):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent3_params = test_suite.get_agent_params('agent3')
    await agent1.open()
    await agent2.open()
    await agent3.open()
    try:
        pw1 = await get_pairwise(agent1, agent2)
        pw2 = await get_pairwise(agent1, agent3)
    finally:
        await agent1.close()
        await agent2.close()
        await agent3.close()

    thread_id = 'thread-id-' + uuid.uuid4().hex
    rcv_messages = []

    async def sender(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal thread_id
        async with sirius_sdk.context(server_address, credentials, p2p):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.CoProtocolThreadedTheirs(thread_id, [pw1, pw2])
            await co.send(msg)

    async def reader(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal rcv_messages
        async with sirius_sdk.context(server_address, credentials, p2p):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                rcv_messages.append(event.message)
                return

    await run_coroutines(
        sender(**agent1_params),
        reader(**agent2_params),
        reader(**agent3_params)
    )

    assert len(rcv_messages) == 2
    msg1 = rcv_messages[0]
    msg2 = rcv_messages[1]
    for msg in [msg1, msg2]:
        assert isinstance(msg1, sirius_sdk.aries_rfc.Ping)
        assert msg['~thread']['thid'] == thread_id
        assert msg.comment == 'Test Ping'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_switch(
        test_suite: ServerTestSuite, agent1: Agent, agent2: Agent, agent3: Agent
):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent3_params = test_suite.get_agent_params('agent3')
    await agent1.open()
    await agent2.open()
    await agent3.open()
    try:
        pw1 = await get_pairwise(agent1, agent2)
        pw2 = await get_pairwise(agent1, agent3)
    finally:
        await agent1.close()
        await agent2.close()
        await agent3.close()

    thread_id = 'thread-id-' + uuid.uuid4().hex
    statuses = None

    async def actor(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal thread_id
        nonlocal statuses
        async with sirius_sdk.context(server_address, credentials, p2p):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.CoProtocolThreadedTheirs(thread_id, [pw1, pw2])
            statuses = await co.switch(msg)

    async def responder(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        async with sirius_sdk.context(server_address, credentials, p2p):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                thread_id_ = event.message['~thread']['thid']
                pong = sirius_sdk.aries_rfc.Pong(ping_id=thread_id_, comment='PONG')
                await sirius_sdk.send_to(pong, event.pairwise)
                return

    await run_coroutines(
        actor(**agent1_params),
        responder(**agent2_params),
        responder(**agent3_params),
    )

    assert statuses is not None
    assert pw1 in statuses.keys()
    assert pw2 in statuses.keys()
    for pw, stat in statuses.items():
        success, message = stat
        assert success is True
        assert message['comment'] == 'PONG'


@pytest.mark.asyncio
async def test_coprotocol_threaded_theirs_switch_timeout(
        test_suite: ServerTestSuite, agent1: Agent, agent2: Agent, agent3: Agent
):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    await agent1.open()
    await agent2.open()
    await agent3.open()
    try:
        pw1 = await get_pairwise(agent1, agent2)
        pw2 = await get_pairwise(agent1, agent3)
    finally:
        await agent1.close()
        await agent2.close()
        await agent3.close()

    ttl = 10
    thread_id = 'thread-id-' + uuid.uuid4().hex
    statuses = None

    async def actor(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        nonlocal thread_id
        nonlocal statuses
        nonlocal ttl
        async with sirius_sdk.context(server_address, credentials, p2p):
            msg = sirius_sdk.aries_rfc.Ping(comment='Test Ping')
            co = sirius_sdk.CoProtocolThreadedTheirs(thread_id, [pw1, pw2], time_to_live=ttl)
            statuses = await co.switch(msg)

    async def responder(server_address: str, credentials: bytes, p2p: sirius_sdk.P2PConnection, **kwargs):
        async with sirius_sdk.context(server_address, credentials, p2p):
            listener = await sirius_sdk.subscribe()
            async for event in listener:
                thread_id_ = event.message['~thread']['thid']
                pong = sirius_sdk.aries_rfc.Pong(ping_id=thread_id_, comment='PONG')
                await sirius_sdk.send_to(pong, event.pairwise)
                return

    await run_coroutines(
        actor(**agent1_params),
        responder(**agent2_params),
        timeout=2*ttl
    )

    assert statuses is not None
    stat1 = statuses[pw1]
    assert stat1[0] is True
    assert stat1[1]['comment'] == 'PONG'
    stat1 = statuses[pw2]
    assert stat1[0] is False
    assert stat1[1] is None



import base64
import datetime

import pytest

from sirius_sdk.messaging import Message
from sirius_sdk.rpc import Future
from sirius_sdk.rpc.futures import MSG_TYPE as MSG_TYPE_FUTURE
from sirius_sdk.errors.exceptions import *
from sirius_sdk.errors.indy_exceptions import WalletItemAlreadyExists, ErrorCode


@pytest.mark.asyncio
async def test_sane(p2p: dict):
    agent_to_sdk = p2p['agent']['tunnel']
    sdk_to_agent = p2p['sdk']['tunnel']

    future = Future(tunnel=sdk_to_agent)

    with pytest.raises(SiriusPendingOperation):
        future.get_value()

    expected = 'Test OK'
    promise_msg = Message({
        '@type': MSG_TYPE_FUTURE,
        '@id': 'promise-message-id',
        'is_tuple': False,
        'is_bytes': False,
        'value': expected,
        'exception': None,
        '~thread': {
            'thid': future.promise['id']
        }
    })

    ok = await future.wait(5)
    assert ok is False

    await agent_to_sdk.post(message=promise_msg)
    ok = await future.wait(5)
    assert ok is True

    actual = future.get_value()
    assert actual == expected


@pytest.mark.asyncio
async def test_set_non_indy_error(p2p: dict):
    agent_to_sdk = p2p['agent']['tunnel']
    sdk_to_agent = p2p['sdk']['tunnel']

    future = Future(tunnel=sdk_to_agent)

    exc = RuntimeError('test error message')

    promise_msg = Message({
        '@type': MSG_TYPE_FUTURE,
        '@id': 'promise-message-id',
        'is_tuple': False,
        'is_bytes': False,
        'value': None,
        'exception': {
            'indy': None,
            'class_name': exc.__class__.__name__,
            'printable': str(exc)
        },
        '~thread': {
            'thid': future.promise['id']
        }
    })

    await agent_to_sdk.post(message=promise_msg)
    ok = await future.wait(5)
    assert ok is True

    has_exc = future.has_exception()
    assert has_exc is True

    fut_exc = None
    try:
        future.raise_exception()
    except SiriusPromiseContextException as exc:
        fut_exc = exc

    assert fut_exc is not None
    assert isinstance(fut_exc, SiriusPromiseContextException)
    assert fut_exc.printable == 'test error message'
    assert fut_exc.class_name == 'RuntimeError'


@pytest.mark.asyncio
async def test_set_indy_error(p2p: dict):
    agent_to_sdk = p2p['agent']['tunnel']
    sdk_to_agent = p2p['sdk']['tunnel']

    future = Future(tunnel=sdk_to_agent)

    exc = WalletItemAlreadyExists(
        error_code=ErrorCode.WalletItemAlreadyExists,
        error_details=dict(message='test error message', indy_backtrace='')
    )

    promise_msg = Message({
        '@type': MSG_TYPE_FUTURE,
        '@id': 'promise-message-id',
        'is_tuple': False,
        'is_bytes': False,
        'value': None,
        'exception': {
            'indy': {
                'error_code': exc.error_code,
                'message': exc.message
            },
            'class_name': exc.__class__.__name__,
            'printable': str(exc)
        },
        '~thread': {
            'thid': future.promise['id']
        }
    })

    await agent_to_sdk.post(message=promise_msg)
    ok = await future.wait(5)
    assert ok is True

    has_exc = future.has_exception()
    assert has_exc is True

    fut_exc = None
    try:
        future.raise_exception()
    except WalletItemAlreadyExists as exc:
        fut_exc = exc

    assert fut_exc is not None
    assert isinstance(fut_exc, WalletItemAlreadyExists)
    assert fut_exc.message == 'test error message'


@pytest.mark.asyncio
async def test_tuple_value(p2p: dict):
    agent_to_sdk = p2p['agent']['tunnel']
    sdk_to_agent = p2p['sdk']['tunnel']

    future = Future(tunnel=sdk_to_agent)

    expected = (1, 2, 'value')
    promise_msg = Message({
        '@type': MSG_TYPE_FUTURE,
        '@id': 'promise-message-id',
        'is_tuple': True,
        'is_bytes': False,
        'value': expected,
        'exception': None,
        '~thread': {
            'thid': future.promise['id']
        }
    })

    await agent_to_sdk.post(message=promise_msg)
    ok = await future.wait(3)
    assert ok is True
    actual = future.get_value()
    assert expected == actual


@pytest.mark.asyncio
async def test_bytes_value(p2p: dict):
    agent_to_sdk = p2p['agent']['tunnel']
    sdk_to_agent = p2p['sdk']['tunnel']

    future = Future(tunnel=sdk_to_agent)

    expected = b'Hello!'
    promise_msg = Message({
        '@type': MSG_TYPE_FUTURE,
        '@id': 'promise-message-id',
        'is_tuple': False,
        'is_bytes': True,
        'value': base64.b64encode(expected).decode('ascii'),
        'exception': None,
        '~thread': {
            'thid': future.promise['id']
        }
    })

    await agent_to_sdk.post(message=promise_msg)
    ok = await future.wait(3)
    assert ok is True
    actual = future.get_value()
    assert expected == actual



from sirius_sdk.agent.aries_rfc.feature_0113_question_answer.messages import Question, Answer
from sirius_sdk.agent.aries_rfc.feature_0113_question_answer.recipes import ask_and_wait_answer, make_answer

__all__ = ['Question', 'Answer', 'ask_and_wait_answer', 'make_answer']



from sirius_sdk.storages.impl.kv_storage import InMemoryKeyValueStorage
from sirius_sdk.storages.impl.immutable import InMemoryImmutableCollection
from sirius_sdk.storages.abstract.immutable import AbstractImmutableCollection
from sirius_sdk.storages.abstract.kv_storage import AbstractKeyValueStorage

__all__ = [
    "InMemoryKeyValueStorage", "InMemoryImmutableCollection", "AbstractImmutableCollection", "AbstractKeyValueStorage"
]



import json
import base64
from typing import List, Optional
from collections import UserDict

from sirius_sdk.errors.exceptions import *
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, AriesProblemReport, THREAD_DECORATOR


CREDENTIAL_PREVIEW_TYPE = "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/issue-credential/1.0/credential-preview"
CREDENTIAL_TRANSLATION_TYPE = "https://github.com/Sirius-social/agent/tree/master/messages/credential-translation"
ISSUER_SCHEMA_TYPE = "https://github.com/Sirius-social/agent/tree/master/messages/issuer-schema"
CREDENTIAL_TRANSLATION_ID = "credential-translation"
ISSUER_SCHEMA_ID = "issuer-schema"


class ProposedAttrib(UserDict):

    def __init__(self, name: str, value: str, mime_type: str = None, **kwargs):
        super().__init__()
        self.data['name'] = name
        if mime_type:
            self.data['mime-type'] = mime_type
        self.data['value'] = value

    def to_json(self):
        return self.data


class AttribTranslation(UserDict):

    def __init__(self, attrib_name: str, translation: str, **kwargs):
        super().__init__()
        self.data['attrib_name'] = attrib_name
        self.data['translation'] = translation

    def to_json(self):
        return self.data


class BaseIssueCredentialMessage(AriesProtocolMessage, metaclass=RegisterMessage):

    PROTOCOL = 'issue-credential'
    DEF_LOCALE = 'en'

    def __init__(self, locale: str = DEF_LOCALE, *args, **kwargs):
        version = kwargs.pop('version', '1.1')
        super().__init__(version=version, *args, **kwargs)
        self['~l10n'] = {"locale": locale}

    @property
    def locale(self) -> str:
        return self.get('~l10n', {}).get('locale', self.DEF_LOCALE)

    @property
    def ack_message_id(self) -> str:
        return self.get('~please_ack', {}).get('message_id', None) or self.id

    @property
    def please_ack(self) -> bool:
        """https://github.com/hyperledger/aries-rfcs/tree/master/features/0317-please-ack"""
        return self.get('~please_ack', None) is not None

    @please_ack.setter
    def please_ack(self, flag: bool):
        if flag:
            self['~please_ack'] = {'message_id': self.id}
        elif '~please_ack' in self:
            del self['~please_ack']

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @thread_id.setter
    def thread_id(self, thid: str):
        thread = self.get(THREAD_DECORATOR, {})
        thread['thid'] = thid
        self[THREAD_DECORATOR] = thread


class IssueProblemReport(AriesProblemReport, metaclass=RegisterMessage):

    PROTOCOL = BaseIssueCredentialMessage.PROTOCOL


class ProposeCredentialMessage(BaseIssueCredentialMessage):

    NAME = 'propose-credential'

    def __init__(
            self, comment: str = None, proposal_attrib: List[ProposedAttrib] = None, schema_id: str = None,
            schema_name: str = None, schema_version: str = None, schema_issuer_did: str = None, cred_def_id: str = None,
            issuer_did: str = None, proposal_attrib_translation: List[AttribTranslation] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        if schema_id:
            self['schema_id'] = schema_id
        if schema_name:
            self['schema_name'] = schema_name
        if schema_version:
            self['schema_version'] = schema_version
        if schema_issuer_did:
            self['schema_issuer_did'] = schema_issuer_did
        if cred_def_id:
            self['cred_def_id'] = cred_def_id
        if issuer_did:
            self['issuer_did'] = issuer_did
        if proposal_attrib:
            self['credential_proposal'] = {
                "@type": CREDENTIAL_PREVIEW_TYPE,
                "attributes": [attrib.to_json() for attrib in proposal_attrib]
            }
            if proposal_attrib_translation:
                self['~attach'] = [
                    {
                        "@type": CREDENTIAL_TRANSLATION_TYPE,
                        "id": CREDENTIAL_TRANSLATION_ID,
                        '~l10n': {"locale": self.locale},
                        "mime-type": "application/json",
                        "data": {
                            "json": [trans.to_json() for trans in proposal_attrib_translation]
                        }
                    }
                ]


class OfferCredentialMessage(BaseIssueCredentialMessage):

    NAME = 'offer-credential'

    def __init__(
            self, comment: str = None, offer: dict = None, cred_def: dict = None,
            preview: List[ProposedAttrib] = None, issuer_schema: dict = None, translation: List[AttribTranslation] = None,
            expires_time: str = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        if preview:
            preview = [ProposedAttrib(**item) for item in preview] if preview else None
            self["credential_preview"] = {
                "@type": CREDENTIAL_PREVIEW_TYPE,
                "attributes": [attrib.to_json() for attrib in preview]
            }
        if translation:
            translation = [AttribTranslation(**item) for item in translation]
        if offer and cred_def:
            payload = dict(**offer, **cred_def)
            self["offers~attach"] = [
                {
                    "@id": 'libindy-cred-offer-' + self.id,
                    "mime-type": "application/json",
                    "data": {
                        "base64": base64.b64encode(json.dumps(payload).encode()).decode()
                    }
                }
            ]
        if translation or issuer_schema:
            self['~attach'] = []
            if translation:
                self['~attach'].append(
                    {
                        "@type": CREDENTIAL_TRANSLATION_TYPE,
                        "id": CREDENTIAL_TRANSLATION_ID,
                        '~l10n': {"locale": self.locale},
                        "mime-type": "application/json",
                        "data": {
                            "json": [trans.to_json() for trans in translation]
                        }
                    }
                )
            if issuer_schema:
                self['~attach'].append(
                    {
                        "@type": ISSUER_SCHEMA_TYPE,
                        "id": ISSUER_SCHEMA_ID,
                        "mime-type": "application/json",
                        "data": {
                            "json": issuer_schema
                        }
                    }
                )
        if expires_time:
            self['~timing'] = {
                "expires_time": expires_time
            }

    @property
    def comment(self) -> Optional[str]:
        return self.get('comment', None)

    @property
    def preview(self) -> Optional[List[ProposedAttrib]]:
        preview = self.get('credential_preview', None)
        if (type(preview) is dict) and (preview.get('@type', None) == CREDENTIAL_PREVIEW_TYPE):
            attribs = preview.get('attributes', [])
            return [ProposedAttrib(**item) for item in attribs]
        else:
            return None

    @property
    def translation(self) -> Optional[List[AttribTranslation]]:
        attaches = self.get('~attach', [])
        tr = None
        for item in attaches:
            if item.get('@type', None) == CREDENTIAL_TRANSLATION_TYPE:
                tr = item
                break
        if tr:
            translation = tr.get('data', {}).get('json', [])
            return [AttribTranslation(**item) for item in translation]
        else:
            return None

    @property
    def issuer_schema(self) -> Optional[dict]:
        attaches = self.get('~attach', [])
        cs = None
        for item in attaches:
            if item.get('@type', None) == ISSUER_SCHEMA_TYPE:
                cs = item
                break
        if cs:
            return cs.get('data', {}).get('json', [])
        else:
            return None

    @property
    def offer(self) -> Optional[dict]:
        try:
            _, offer, _ = self.parse()
        except SiriusValidationError:
            return None
        else:
            return offer

    @property
    def cred_def(self) -> Optional[dict]:
        try:
            _, _, cred_def = self.parse()
        except SiriusValidationError:
            return None
        else:
            return cred_def

    @property
    def expires_time(self) -> Optional[str]:
        return self.get('~timing', {}).get('expires_time', None)

    def parse(self) -> (dict, dict, dict):
        offer_attaches = self.get('offers~attach', None)
        if isinstance(offer_attaches, dict):
            offer_attaches = [offer_attaches]
        if (not type(offer_attaches) is list) or (type(offer_attaches) is list and len(offer_attaches) == 0):
            raise SiriusValidationError('Expected attribute "offer~attach" must contains cred-Offer and cred-Def')
        offer = offer_attaches[0]
        offer_body = None
        cred_def_body = None

        for attach in offer_attaches:
            raw_base64 = attach.get('data', {}).get('base64', None)
            if raw_base64:
                payload = json.loads(base64.b64decode(raw_base64).decode())
                offer_fields = ['key_correctness_proof', 'nonce', 'schema_id', 'cred_def_id']
                cred_def_fields = ['value', 'type', 'ver', 'schemaId', 'id', 'tag']
                if all([field in payload.keys() for field in offer_fields]):  # check if cred offer content
                    offer_body = {attr: val for attr, val in payload.items() if attr in offer_fields}
                if all([field in payload.keys() for field in cred_def_fields]):  # check if cred def content
                    cred_def_body = {attr: val for attr, val in payload.items() if attr in cred_def_fields}

        if not offer_body:
            raise SiriusValidationError('Expected offer~attach must contains Payload with offer')

        if not cred_def_body:
            raise SiriusValidationError('Expected offer~attach must contains Payload with cred_def data')

        return offer, offer_body, cred_def_body

    def validate(self):
        super().validate()
        if 'offers~attach' not in self:
            raise SiriusValidationError('Expected offer attribute "offers~attach" missing')
        self.parse()


class RequestCredentialMessage(BaseIssueCredentialMessage):

    NAME = 'request-credential'

    def __init__(
            self, comment: str = None, cred_request: dict = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        if cred_request:
            self["requests~attach"] = [
                {
                    "@id": 'cred-request-' + self.id,
                    "mime-type": "application/json",
                    "data": {
                        "base64": base64.b64encode(json.dumps(cred_request).encode()).decode()
                    }
                },
            ]

    @property
    def cred_request(self) -> Optional[dict]:
        request = self.get('requests~attach', None)
        if request:
            if isinstance(request, list):
                request = request[0]
            body = request.get('data').get('base64')
            body = base64.b64decode(body)
            body = json.loads(body.decode())
            return body
        else:
            return None

    def validate(self):
        super().validate()
        if 'requests~attach' not in self:
            raise SiriusValidationError('Expected offer attribute "requests~attach" missing')


class IssueCredentialMessage(BaseIssueCredentialMessage):

    NAME = 'issue-credential'

    def __init__(
            self, comment: str = None, cred: dict = None, cred_id: str = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if comment:
            self['comment'] = comment
        if cred:
            if cred_id:
                message_id = cred_id
            else:
                message_id = 'libindy-cred-' + self.id
            self["credentials~attach"] = [
                {
                    "@id": message_id,
                    "mime-type": "application/json",
                    "data": {
                        "base64": base64.b64encode(json.dumps(cred).encode()).decode()
                    }
                }
            ]

    @property
    def cred_id(self) -> Optional[str]:
        attaches = self.get('credentials~attach', None)
        if attaches:
            if isinstance(attaches, dict):
                attaches = [attaches]
            if isinstance(attaches, list):
                attach = attaches[0]
                return attach.get('@id', None)
            else:
                return None
        else:
            return None

    @property
    def cred(self) -> Optional[dict]:
        attaches = self.get('credentials~attach', None)
        if attaches:
            if isinstance(attaches, dict):
                attaches = [attaches]
            if isinstance(attaches, list):
                attach = attaches[0]
                b64 = attach.get('data', {}).get('base64', None)
                if b64:
                    body = base64.b64decode(b64)
                    body = json.loads(body.decode())
                    return body
                else:
                    return None
            else:
                return None
        else:
            return None

    def validate(self):
        super().validate()
        if 'credentials~attach' not in self:
            raise SiriusValidationError('Expected issue attribute "credentials~attach" missing')
        if self.cred is None:
            raise SiriusValidationError('Credential is empty in "credentials~attach" field')



from typing import List

from sirius_sdk.hub import acquire, release


class Locking:

    NAMESPACE = 'ledgers'

    @classmethod
    async def acquire(cls, names: List[str], lock_timeout: float) -> (bool, List[str]):
        """Lock ledgers given by names.

        :names: names of microledgers
        :lock_timeout: lock timeout, resources will be released automatically after timeout expired
        """
        ledger_names = names[:]  # copy
        ledger_names = list(set(ledger_names))  # remove duplicates
        ledger_resources = [f'{cls.NAMESPACE}/{name}' for name in ledger_names]
        ok, locked_ledgers = await acquire(resources=ledger_resources, lock_timeout=lock_timeout)
        locked_ledgers = [item.split('/')[-1] for item in locked_ledgers]  # remove namespace prefix
        return ok, locked_ledgers
        pass

    @classmethod
    async def release(cls):
        """Released all resources locked in current context"""
        await release()
        pass


__all__ = ['Locking']



from typing import Optional, List, Union

from sirius_sdk.agent.pairwise import AbstractPairwiseList
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache
from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.messaging import Message
from sirius_sdk.agent.ledger import Ledger
from sirius_sdk.agent.listener import Listener
from sirius_sdk.agent.connections import Endpoint
from sirius_sdk.agent.microledgers.abstract import AbstractMicroledgerList

from .core import _current_hub, init, context
from .proxies import DIDProxy, CryptoProxy, MicroledgersProxy, PairwiseProxy, AnonCredsProxy, \
    CacheProxy, NonSecretsProxy
from .coprotocols import CoProtocolThreadedP2P, CoProtocolP2PAnon, CoProtocolP2P, AbstractP2PCoProtocol, \
    CoProtocolThreadedTheirs, open_communication

DID: AbstractDID = DIDProxy()
Crypto: AbstractCrypto = CryptoProxy()
Microledgers: AbstractMicroledgerList = MicroledgersProxy()
PairwiseList: AbstractPairwiseList = PairwiseProxy()
AnonCreds: AnonCredsProxy = AnonCredsProxy()
Cache: AbstractCache = CacheProxy()
NonSecrets: AbstractNonSecrets = NonSecretsProxy()


async def ledger(name: str) -> Optional[Ledger]:
    async with _current_hub().get_agent_connection_lazy() as agent:
        return agent.ledger(name)


async def endpoints() -> List[Endpoint]:
    async with _current_hub().get_agent_connection_lazy() as agent:
        return agent.endpoints


async def subscribe() -> Listener:
    async with _current_hub().get_agent_connection_lazy() as agent:
        return await agent.subscribe()


async def ping() -> bool:
    async with _current_hub().get_agent_connection_lazy() as agent:
        return await agent.ping()


async def send(
        message: Message, their_vk: Union[List[str], str],
        endpoint: str, my_vk: Optional[str], routing_keys: Optional[List[str]] = None
):
    async with _current_hub().get_agent_connection_lazy() as agent:
        await agent.send_message(
            message=message, their_vk=their_vk,
            endpoint=endpoint, my_vk=my_vk, routing_keys=routing_keys
        )


async def send_to(message: Message, to: Pairwise):
    async with _current_hub().get_agent_connection_lazy() as agent:
        await agent.send_to(message=message, to=to)


async def generate_qr_code(value: str) -> str:
    async with _current_hub().get_agent_connection_lazy() as agent:
        return await agent.generate_qr_code(value)


async def acquire(resources: List[str], lock_timeout: float, enter_timeout: float = None) -> (bool, List[str]):
    async with _current_hub().get_agent_connection_lazy() as agent:
        kwargs = {'resources': resources, 'lock_timeout': lock_timeout}
        if enter_timeout:
            kwargs['enter_timeout'] = enter_timeout
        return await agent.acquire(**kwargs)


async def release():
    async with _current_hub().get_agent_connection_lazy() as agent:
        await agent.release()



import json
import copy
import logging
from typing import List, Optional, Union

from sirius_sdk.base import JsonSerializable
from sirius_sdk.storages import AbstractImmutableCollection
from sirius_sdk.errors.indy_exceptions import LedgerNotFound
from sirius_sdk.errors.exceptions import SiriusInvalidPayloadStructure
from sirius_sdk.agent.wallet.abstract.ledger import AbstractLedger, NYMRole
from sirius_sdk.agent.wallet.abstract.anoncreds import AnonCredSchema, AbstractAnonCreds
from sirius_sdk.agent.wallet.abstract.cache import AbstractCache, CacheOptions


class Schema(AnonCredSchema, JsonSerializable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def seq_no(self) -> int:
        return self.body.get('seqNo', None)

    @property
    def issuer_did(self) -> str:
        parts = self.id.split(':')
        return parts[0]

    def __eq__(self, other):
        if isinstance(other, Schema):
            equal = super().__eq__(other)
            if equal:
                return self.seq_no == other.seq_no
            else:
                return False
        else:
            return False

    def serialize(self) -> dict:
        return self.body

    @classmethod
    def deserialize(cls, buffer: Union[dict, bytes, str]):
        if isinstance(buffer, bytes):
            kwargs = json.loads(buffer.decode())
        elif isinstance(buffer, str):
            kwargs = json.loads(buffer)
        elif isinstance(buffer, dict):
            kwargs = buffer
        else:
            raise RuntimeError('Unexpected buffer Type')
        return Schema(**kwargs)


class SchemaFilters:

    def __init__(self):
        self.__tags = {'category': 'schema'}

    @property
    def tags(self) -> dict:
        return self.__tags

    @property
    def id(self) -> Optional[str]:
        return self.__tags.get('id', None)

    @id.setter
    def id(self, value: str):
        self.__tags['id'] = value

    @property
    def name(self) -> Optional[str]:
        return self.__tags.get('name', None)

    @name.setter
    def name(self, value: str):
        self.__tags['name'] = value

    @property
    def version(self) -> Optional[str]:
        return self.__tags.get('version', None)

    @version.setter
    def version(self, value: str):
        self.__tags['version'] = value

    @property
    def submitter_did(self) -> Optional[str]:
        return self.__tags.get('submitter_did', None)

    @submitter_did.setter
    def submitter_did(self, value: str):
        self.__tags['submitter_did'] = value


class CredentialDefinition(JsonSerializable):

    class Config(JsonSerializable):

        def __init__(self):
            self.__support_revocation = False

        @property
        def support_revocation(self) -> bool:
            return self.__support_revocation

        @support_revocation.setter
        def support_revocation(self, value: bool):
            self.__support_revocation = value

        def serialize(self) -> dict:
            return {
                'support_revocation': self.support_revocation
            }

        @classmethod
        def deserialize(cls, buffer: Union[dict, bytes, str]):
            if isinstance(buffer, bytes):
                data = json.loads(buffer.decode())
            elif isinstance(buffer, str):
                data = json.loads(buffer)
            elif isinstance(buffer, dict):
                data = buffer
            else:
                raise RuntimeError('Unexpected buffer Type')
            instance = CredentialDefinition.Config()
            instance.support_revocation = data.get('support_revocation', False)
            return instance

    def __init__(self, tag: str, schema: Schema, config: Config = None, body: dict = None, seq_no: int = None):
        self.__tag = tag
        self.__schema = schema
        self.__config = config or CredentialDefinition.Config()
        self.__body = body
        self.__seq_no = seq_no

    @property
    def tag(self) -> str:
        return self.__tag

    @property
    def id(self) -> Optional[str]:
        if self.body:
            return self.body.get('id', None)
        else:
            return None

    @property
    def submitter_did(self) -> Optional[str]:
        if self.id:
            parts = self.id.split(':')
            return parts[0]
        else:
            return None

    @property
    def seq_no(self) -> Optional[int]:
        return self.__seq_no

    @property
    def schema(self) -> Schema:
        return self.__schema

    @property
    def config(self) -> Config:
        return self.__config

    @property
    def body(self) -> dict:
        return self.__body

    def serialize(self) -> dict:
        return {
            'schema': self.schema.serialize(),
            'config': self.config.serialize(),
            'body': self.body,
            'seq_no': self.seq_no
        }

    @classmethod
    def deserialize(cls, buffer: Union[dict, bytes, str]):
        if isinstance(buffer, bytes):
            data = json.loads(buffer.decode())
        elif isinstance(buffer, str):
            data = json.loads(buffer)
        elif isinstance(buffer, dict):
            data = buffer
        else:
            raise RuntimeError('Unexpected buffer Type')
        schema = Schema.deserialize(data['schema'])
        config = CredentialDefinition.Config.deserialize(data['config'])
        body = data['body']
        seq_no = data['seq_no']
        return CredentialDefinition(body['tag'], schema, config, body, seq_no)


class CredentialDefinitionFilters:

    def __init__(self):
        self.__extras = {}
        self.__tags = {'category': 'cred_def'}

    @property
    def tags(self) -> dict:
        d = copy.copy(self.__tags)
        d.update(self.__extras)
        return d

    @property
    def extras(self) -> dict:
        return self.__extras

    @extras.setter
    def extras(self, value: dict):
        self.__extras = value

    def extra(self, name: str, value: str):
        self.__extras[name] = value

    @property
    def tag(self) -> Optional[str]:
        return self.__tags.get('tag', None)

    @tag.setter
    def tag(self, value: str):
        if value:
            self.__tags['tag'] = value
        elif 'tag' in self.__tags:
            del self.__tags['tag']

    @property
    def id(self) -> Optional[str]:
        return self.__tags.get('id', None)

    @id.setter
    def id(self, value: str):
        self.__tags['id'] = value

    @property
    def submitter_did(self) -> Optional[str]:
        return self.__tags.get('submitter_did', None)

    @submitter_did.setter
    def submitter_did(self, value: str):
        self.__tags['submitter_did'] = value

    @property
    def schema_id(self) -> Optional[str]:
        return self.__tags.get('schema_id', None)

    @schema_id.setter
    def schema_id(self, value: str):
        self.__tags['schema_id'] = value

    @property
    def seq_no(self) -> Optional[int]:
        value = self.__tags.get('seq_no', None)
        return int(value) if value else None

    @seq_no.setter
    def seq_no(self, value: int):
        self.__tags['seq_no'] = str(value)


class Ledger:

    def __init__(
            self, name: str, api: AbstractLedger, issuer: AbstractAnonCreds,
            cache: AbstractCache, storage: AbstractImmutableCollection
    ):
        self.__name = name
        self._api = api
        self._cache = cache
        self._issuer = issuer
        self._storage = storage
        self.__db = 'ledger_storage_%s' % name

    @property
    def name(self) -> str:
        return self.__name

    async def read_nym(self, submitter_did: str, target_did: str) -> (bool, dict):
        resp = await self._api.read_nym(
            pool_name=self.name,
            submitter_did=submitter_did,
            target_did=target_did
        )
        success = resp[0]
        data = resp[1]
        return success, data

    async def write_nym(
            self, submitter_did: str, target_did: str, ver_key: str = None, alias: str = None, role: NYMRole = None
    ) -> (bool, dict):
        resp = await self._api.write_nym(
            pool_name=self.name,
            submitter_did=submitter_did,
            target_did=target_did,
            ver_key=ver_key,
            alias=alias,
            role=role or NYMRole.COMMON_USER
        )
        success = resp[0]
        data = resp[1]
        return success, data

    async def load_schema(self, id_: str, submitter_did: str) -> Schema:
        body = await self._cache.get_schema(
            pool_name=self.name,
            submitter_did=submitter_did,
            id_=id_,
            options=CacheOptions()
        )
        return Schema(**body)

    async def load_cred_def(self, id_: str, submitter_did: str) -> CredentialDefinition:
        cred_def_body = await self._cache.get_cred_def(
            pool_name=self.name,
            submitter_did=submitter_did,
            id_=id_,
            options=CacheOptions()
        )
        tag = cred_def_body.get('tag')
        schema_seq_no = int(cred_def_body['schemaId'])
        cred_def_seq_no = int(cred_def_body['id'].split(':')[3]) + 1
        txn_request = await self._api.build_get_txn_request(
            submitter_did=submitter_did,
            ledger_type=None,
            seq_no=schema_seq_no
        )
        resp = await self._api.sign_and_submit_request(
            pool_name=self.name,
            submitter_did=submitter_did,
            request=txn_request
        )
        if resp['op'] == 'REPLY':
            txn_data = resp['result']['data']
            schema_body = {
                'name': txn_data['txn']['data']['data']['name'],
                'version': txn_data['txn']['data']['data']['version'],
                'attrNames': txn_data['txn']['data']['data']['attr_names'],
                'id': txn_data['txnMetadata']['txnId'],
                'seqNo': txn_data['txnMetadata']['seqNo']
            }
            schema_body['ver'] = schema_body['id'].split(':')[-1]
            schema = Schema(**schema_body)
            cred_def = CredentialDefinition(
                tag=tag, schema=schema, body=cred_def_body, seq_no=cred_def_seq_no
            )
            return cred_def
        else:
            raise SiriusInvalidPayloadStructure()

    async def register_schema(self, schema: AnonCredSchema, submitter_did: str) -> (bool, Schema):
        success, txn_response = await self._api.register_schema(
            pool_name=self.name,
            submitter_did=submitter_did,
            data=schema.body
        )
        if success and txn_response.get('op') == 'REPLY':
            body = copy.copy(schema.body)
            body['seqNo'] = txn_response['result']['txnMetadata']['seqNo']
            schema_in_ledger = Schema(**body)
            await self.__ensure_exists_in_storage(schema_in_ledger, submitter_did)
            return True, schema_in_ledger
        else:
            reason = txn_response.get('reason', None)
            if reason:
                logging.error(reason)
            return False, None

    async def register_cred_def(
            self, cred_def: CredentialDefinition, submitter_did: str, tags: dict = None
    ) -> (bool, CredentialDefinition):
        cred_def_id, body = await self._issuer.issuer_create_and_store_credential_def(
            issuer_did=submitter_did,
            schema=cred_def.schema.body,
            tag=cred_def.tag,
            config=cred_def.config.serialize()
        )
        build_request = await self._api.build_cred_def_request(
            submitter_did=submitter_did,
            data=body
        )
        signed_request = await self._api.sign_request(
            submitter_did=submitter_did,
            request=build_request
        )
        resp = await self._api.submit_request(self.name, signed_request)
        success = resp.get('op', None) == 'REPLY'
        if success:
            txn_response = resp
        else:
            return False, None
        if success:
            ledger_cred_def = CredentialDefinition(
                tag=cred_def.tag,
                schema=cred_def.schema,
                config=cred_def.config,
                body=body,
                seq_no=txn_response['result']['txnMetadata']['seqNo']
            )
            await self.__ensure_exists_in_storage(ledger_cred_def, submitter_did, tags)
            return True, ledger_cred_def

    async def ensure_schema_exists(self, schema: AnonCredSchema, submitter_did: str) -> Optional[Schema]:
        try:
            body = await self._cache.get_schema(
                pool_name=self.name,
                submitter_did=submitter_did,
                id_=schema.id,
                options=CacheOptions()
            )
            ledger_schema = Schema(**body)
            await self.__ensure_exists_in_storage(ledger_schema, submitter_did)
            return ledger_schema
        except LedgerNotFound:
            pass
        ok, ledger_schema = await self.register_schema(schema, submitter_did)
        if ok:
            return ledger_schema
        else:
            return None

    async def fetch_schemas(
            self, id_: str = None, name: str = None, version: str = None, submitter_did: str = None
    ) -> List[Schema]:
        filters = SchemaFilters()
        if id_:
            filters.id = id_
        if name:
            filters.name = name
        if version:
            filters.version = version
        if submitter_did:
            filters.submitter_did = submitter_did
        await self._storage.select_db(self.__db)
        fetched, total_count = await self._storage.fetch(filters.tags)
        return [Schema.deserialize(item) for item in fetched]

    async def fetch_cred_defs(
            self, tag: str = None, id_: str = None, submitter_did: str = None,
            schema_id: str = None, seq_no: int = None, **kwargs
    ) -> List[CredentialDefinition]:
        filters = CredentialDefinitionFilters()
        if tag:
            filters.tag = tag
        if id_:
            filters.id = id_
        if submitter_did:
            filters.submitter_did = submitter_did
        if schema_id:
            filters.schema_id = schema_id
        if seq_no:
            filters.seq_no = seq_no
        filters.extras = kwargs
        await self._storage.select_db(self.__db)
        fetched, total_count = await self._storage.fetch(filters.tags)
        return [CredentialDefinition.deserialize(item) for item in fetched]

    async def __ensure_exists_in_storage(
            self, entity: Union[Schema, CredentialDefinition], submitter_did: str, search_tags: dict = None
    ):
        await self._storage.select_db(self.__db)
        if isinstance(entity, Schema):
            schema = entity
            tags = {
                'id': schema.id,
                'category': 'schema'
            }
            _, count = await self._storage.fetch(tags=tags)
            if count == 0:
                tags.update(
                    {
                        'id': schema.id,
                        'name': schema.name,
                        'version': schema.version,
                        'submitter_did': submitter_did
                    }
                )
                await self._storage.add(
                    value=schema.serialize(),
                    tags=tags
                )
        elif isinstance(entity, CredentialDefinition):
            cred_def = entity
            tags = {
                'id': cred_def.id,
                'seq_no': str(cred_def.seq_no),
                'category': 'cred_def'
            }
            _, count = await self._storage.fetch(tags=tags)
            if count == 0:
                tags.update(
                    {
                        'id': cred_def.id,
                        'tag': cred_def.tag,
                        'schema_id': cred_def.schema.id,
                        'submitter_did': cred_def.submitter_did
                    }
                )
                if search_tags:
                    tags.update(search_tags)
                await self._storage.add(
                    value=cred_def.serialize(),
                    tags=tags
                )



import uuid
import asyncio

import pytest

from sirius_sdk import Agent
from sirius_sdk.messaging import Message, register_message_class
from .helpers import ServerTestSuite


@pytest.mark.asyncio
async def test_same_account(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    session1 = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    session2 = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    await session1.open()
    await session2.open()
    try:
        # check locking OK
        resources = [f'resource-{uuid.uuid4().hex}' for i in range(100)]
        ok, busy = await session1.acquire(resources=resources, lock_timeout=5)
        try:
            assert ok is True
            ok, busy = await session2.acquire(resources=resources, lock_timeout=1)
            assert ok is False
            assert set(busy) == set(resources)
        finally:
            await session1.release()
        # check session ok may lock after explicitly release
        ok, busy = await session2.acquire(resources=resources, lock_timeout=1)
        assert ok is True
        # Check after timeout
        resources = [f'resource-{uuid.uuid4().hex}' for i in range(100)]
        timeout = 5.0
        ok, _ = await session1.acquire(resources=resources, lock_timeout=timeout)
        assert ok is True
        ok, _ = await session2.acquire(resources=resources, lock_timeout=1.0)
        assert ok is False
        await asyncio.sleep(timeout + 1.0)
        ok, _ = await session2.acquire(resources=resources, lock_timeout=1.0)
        assert ok is True
    finally:
        await session1.close()
        await session2.close()


@pytest.mark.asyncio
async def test_lock_multiple_time(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    session1 = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    session2 = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    await session1.open()
    await session2.open()
    try:
        resources1 = [f'resource-{uuid.uuid4().hex}']
        timeout = 5.0
        ok, _ = await session1.acquire(resources=resources1, lock_timeout=timeout)
        assert ok is True

        resources2 = [f'resource-{uuid.uuid4().hex}']
        ok, _ = await session1.acquire(resources=resources2, lock_timeout=timeout)
        assert ok is True
        # session1 must unlock previously locked resources on new acquire call
        ok, _ = await session2.acquire(resources=resources1, lock_timeout=timeout)
        assert ok is True
    finally:
        await session1.close()
        await session2.close()


@pytest.mark.asyncio
async def test_different_accounts(test_suite: ServerTestSuite):
    params1 = test_suite.get_agent_params('agent1')
    params2 = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=params1['server_address'],
        credentials=params1['credentials'],
        p2p=params1['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=params2['server_address'],
        credentials=params2['credentials'],
        p2p=params2['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        same_resources = [f'resource/{uuid.uuid4().hex}']

        ok1, _ = await agent1.acquire(resources=same_resources, lock_timeout=10.0)
        ok2, _ = await agent1.acquire(resources=same_resources, lock_timeout=10.0)

        assert ok1 is True
        assert ok2 is True
    finally:
        await agent1.close()
        await agent2.close()



r"""https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol
"""
import re
import base64
from typing import List, Optional
from urllib.parse import urljoin

from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging import Message, check_for_attributes
from sirius_sdk.agent.agent import Agent
from sirius_sdk.agent.wallet.abstract.crypto import AbstractCrypto
from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, AriesProblemReport, THREAD_DECORATOR
from sirius_sdk.agent.aries_rfc.did_doc import DIDDoc
from sirius_sdk.agent.aries_rfc.utils import sign, verify_signed


class ConnProtocolMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries feature 0160 Message implementation

    https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol
    """

    PROTOCOL = 'connections'

    @staticmethod
    async def sign_field(crypto: AbstractCrypto, field_value: Any, my_verkey: str) -> dict:
        return await sign(crypto, field_value, my_verkey)

    @staticmethod
    async def verify_signed_field(crypto: AbstractCrypto, signed_field: dict) -> (Any, bool):
        return await verify_signed(crypto, signed_field)

    @staticmethod
    def build_did_doc(did: str, verkey: str, endpoint: str, **extra):
        key_id = did + '#1'
        doc = {
            "@context": "https://w3id.org/did/v1",
            "id": did,
            "authentication": [
                {
                    "publicKey": key_id,
                    "type": "Ed25519SignatureAuthentication2018"
                }
            ],
            "publicKey": [{
                "id": "1",
                "type": "Ed25519VerificationKey2018",
                "controller": did,
                "publicKeyBase58": verkey
            }],
            "service": [{
                "id": 'did:peer:' + did + ";indy",
                "type": "IndyAgent",
                "priority": 0,
                "recipientKeys": [key_id],
                "serviceEndpoint": endpoint,
            }],
        }
        doc.update(extra)
        return doc

    @property
    def their_did(self):
        return self['connection'].get('did', None) or self['connection'].get('DID')

    @property
    def did_doc(self):
        payload = self.get('connection', {}).get('did_doc', {}) or self.get('connection', {}).get('DIDDoc', None)
        return DIDDoc(payload) if payload is not None else None

    def extract_their_info(self):
        """ Extract the other participant's DID, verkey and endpoint

        :param key: attribute for extracting
        :return: Return a 4-tuple of (DID, verkey, endpoint, routingKeys)
        """
        if self.their_did is None:
            raise SiriusInvalidMessage('Connection metadata is empty')
        if self.did_doc is None:
            raise SiriusInvalidMessage('DID Doc is empty')
        service = self.did_doc.extract_service()
        their_endpoint = service['serviceEndpoint']
        public_keys = self.did_doc['publicKey']

        def get_key(controller_: str, id_: str):
            for k in public_keys:
                if k['controller'] == controller_ and k["id"] == id_:
                    return k['publicKeyBase58']
            return None

        def extract_key(name: str):
            if "#" in name:
                controller_, id_ = name.split('#')
                return get_key(controller_, id_)
            else:
                return name

        their_vk = extract_key(service["recipientKeys"][0])

        routing_keys = []
        for rk in service.get("routingKeys", []):
            routing_keys.append(extract_key(rk))

        return self.their_did, their_vk, their_endpoint, routing_keys

    def validate(self):
        super().validate()
        if 'connection' in self:
            if self.did_doc is None:
                raise SiriusInvalidMessage('DIDDoc is empty')
            self.did_doc.validate()

    @property
    def ack_message_id(self) -> str:
        return self.get('~please_ack', {}).get('message_id', None) or self.id

    @property
    def please_ack(self) -> bool:
        """https://github.com/hyperledger/aries-rfcs/tree/master/features/0317-please-ack"""
        return self.get('~please_ack', None) is not None

    @please_ack.setter
    def please_ack(self, flag: bool):
        if flag:
            self['~please_ack'] = {'message_id': self.id}
        elif '~please_ack' in self:
            del self['~please_ack']

    @property
    def thread_id(self) -> Optional[str]:
        return self.get(THREAD_DECORATOR, {}).get('thid', None)

    @thread_id.setter
    def thread_id(self, thid: str):
        thread = self.get(THREAD_DECORATOR, {})
        thread['thid'] = thid
        self[THREAD_DECORATOR] = thread


class ConnProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    PROTOCOL = ConnProtocolMessage.PROTOCOL


class Invitation(ConnProtocolMessage, metaclass=RegisterMessage):

    NAME = 'invitation'

    def __init__(
            self, label: Optional[str] = None, recipient_keys: Optional[List[str]] = None,
            endpoint: Optional[str] = None, routing_keys: Optional[List[str]] = None, did: str = None,  *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if label is not None:
            self['label'] = label
        if recipient_keys is not None:
            self['recipientKeys'] = recipient_keys
        if endpoint is not None:
            self['serviceEndpoint'] = endpoint
        if routing_keys is not None:
            self['routingKeys'] = routing_keys
        if did is not None:
            self['did'] = did

    def validate(self):
        check_for_attributes(
            self,
            ['label', 'recipientKeys', 'serviceEndpoint']
        )

    @classmethod
    def from_url(cls, url: str) -> ConnProtocolMessage:
        matches = re.match("(.+)?c_i=(.+)", url)
        if not matches:
            raise SiriusInvalidMessage("Invite string is improperly formatted")
        msg = Message.deserialize(base64.urlsafe_b64decode(matches.group(2)).decode('utf-8'))
        if msg.protocol != cls.PROTOCOL:
            raise SiriusInvalidMessage('Unexpected protocol "%s"' % msg.type.protocol)
        if msg.name != cls.NAME:
            raise SiriusInvalidMessage('Unexpected protocol name "%s"' % msg.type.name)
        label = msg.pop('label')
        if label is None:
            raise SiriusInvalidMessage('label attribute missing')
        recipient_keys = msg.pop('recipientKeys')
        if recipient_keys is None:
            raise SiriusInvalidMessage('recipientKeys attribute missing')
        endpoint = msg.pop('serviceEndpoint')
        if endpoint is None:
            raise SiriusInvalidMessage('serviceEndpoint attribute missing')
        routing_keys = msg.pop('routingKeys', [])
        return Invitation(label, recipient_keys, endpoint, routing_keys, **msg)

    @property
    def invitation_url(self) -> str:
        b64_invite = base64.urlsafe_b64encode(self.serialize().encode('ascii')).decode('ascii')
        return '?c_i=' + b64_invite

    async def allocate_qr(self, base_url: str, agent: Agent) -> str:
        full_url = urljoin(base_url, self.invitation_url)
        return await agent.generate_qr_code(value=full_url)

    @property
    def label(self) -> str:
        return self.get('label', None)

    @property
    def recipient_keys(self) -> List[str]:
        return self.get('recipientKeys', [])

    @property
    def endpoint(self) -> str:
        return self.get('serviceEndpoint', None)

    @property
    def routing_keys(self) -> List[str]:
        return self.get('routingKeys', [])


class ConnRequest(ConnProtocolMessage, metaclass=RegisterMessage):

    NAME = 'request'

    def __init__(
            self, label: Optional[str] = None, did: Optional[str] = None, verkey: Optional[str] = None,
            endpoint: Optional[str] = None, did_doc_extra: dict = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if label is not None:
            self['label'] = label
        if (did is not None) and (verkey is not None) and (endpoint is not None):
            extra = did_doc_extra or {}
            self['connection'] = {
                'DID': did,
                'DIDDoc': self.build_did_doc(did, verkey, endpoint, **extra)
            }

    @property
    def label(self) -> Optional[str]:
        return self.get('label', None)

    def validate(self):
        super().validate()
        check_for_attributes(
            self,
            ['label', 'connection']
        )


class ConnResponse(ConnProtocolMessage, metaclass=RegisterMessage):

    NAME = 'response'

    def __init__(
            self, did: Optional[str] = None, verkey: Optional[str] = None,
            endpoint: Optional[str] = None, did_doc_extra: dict = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if (did is not None) and (verkey is not None) and (endpoint is not None):
            extra = did_doc_extra or {}
            self['connection'] = {
                'DID': did,
                'DIDDoc': self.build_did_doc(did, verkey, endpoint, **extra)
            }

    def validate(self):
        super().validate()
        check_for_attributes(
            self,
            ['connection~sig']
        )

    async def sign_connection(self, crypto: AbstractCrypto, key: str):
        self['connection~sig'] = \
            await self.sign_field(
                crypto=crypto, field_value=self['connection'], my_verkey=key
            )
        del self['connection']

    async def verify_connection(self, crypto: AbstractCrypto) -> bool:
        connection, success = await self.verify_signed_field(
            crypto=crypto, signed_field=self['connection~sig']
        )
        if success:
            self['connection'] = connection
        return success



""" Define Message base class.

https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0020-message-types
https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0008-message-id-and-threading
"""
import json
import uuid

from sirius_sdk.errors.exceptions import *
from sirius_sdk.messaging.type import Type, Semver


# Registry for restoring message instance from payload
MSG_REGISTRY = {}


def generate_id():
    """ Generate a message id. """
    return str(uuid.uuid4())


class Message(dict):
    """ Message base class.
        Inherits from dict meaning it behaves like a dictionary.
    """
    __slots__ = (
        'mtc',
        '_type'
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if '@type' not in self:
            raise SiriusInvalidMessage('No @type in message')

        if '@id' not in self:
            self['@id'] = generate_id()
        elif not isinstance(self['@id'], str):
            raise SiriusInvalidMessage('Message @id is invalid; must be str')

        if isinstance(self['@type'], Type):
            self._type = self['@type']
            self['@type'] = str(self._type)
        else:
            self._type = Type.from_str(self.type)

    @property
    def type(self):
        """ Shortcut for msg['@type'] """
        return self['@type']

    @property
    def id(self):  # pylint: disable=invalid-name
        """ Shortcut for msg['@id'] """
        return self['@id']

    @property
    def doc_uri(self) -> str:
        """ Get type doc_uri """
        return self._type.doc_uri

    @property
    def protocol(self) -> str:
        """ Get type protocol """
        return self._type.protocol

    @property
    def version(self) -> str:
        """ Get type version """
        return self._type.version

    @property
    def version_info(self) -> Semver:
        """ Get type version info """
        return self._type.version_info

    @property
    def name(self) -> str:
        """ Get type name """
        return self._type.name

    @property
    def normalized_version(self) -> str:
        """ Get type normalized version """
        return str(self._type.version_info)

    # Serialization
    @classmethod
    def deserialize(cls, serialized: str):
        """ Deserialize a message from a json string. """
        try:
            return cls(json.loads(serialized))
        except json.decoder.JSONDecodeError as err:
            raise SiriusInvalidMessage('Could not deserialize message') from err

    def serialize(self):
        """ Serialize a message into a json string. """
        return json.dumps(self)

    def pretty_print(self):
        """ return a 'pretty print' representation of this message. """
        return json.dumps(self, indent=2)

    def __eq__(self, other):
        if not isinstance(other, Message):
            return False

        return super().__eq__(other)

    def __hash__(self):
        return hash(self.id)


def register_message_class(cls, protocol: str, name: str=None):
    if issubclass(cls, Message):
        descriptor = MSG_REGISTRY.get(protocol, {})
        if name:
            descriptor[name] = cls
        else:
            descriptor['*'] = cls
        MSG_REGISTRY[protocol] = descriptor
    else:
        raise SiriusInvalidMessageClass()


def restore_message_instance(payload: dict) -> (bool, Message):
    if '@type' in payload:
        typ = Type.from_str(payload['@type'])
        descriptor = MSG_REGISTRY.get(typ.protocol, None)
        if descriptor:
            if typ.name in descriptor:
                cls = descriptor[typ.name]
            elif '*' in descriptor:
                cls = descriptor['*']
            else:
                cls = None
        else:
            cls = None
        if cls is not None:
            return True, cls(**payload)
        else:
            return False, None
    else:
        return False, None



import uuid
from datetime import datetime

import pytest

from sirius_sdk import Agent
from sirius_sdk.agent.pairwise import Pairwise
from sirius_sdk.messaging import Message
from sirius_sdk.agent.coprotocols import AbstractCoProtocolTransport
from .conftest import get_pairwise
from .helpers import run_coroutines


TEST_ITERATIONS = 100


async def routine_for_pinger(agent: Agent, p: Pairwise, thread_id: str):
    transport = await agent.spawn(thread_id, p)
    await transport.start()
    try:
        for n in range(TEST_ITERATIONS):
            ping = Message({
                '@id': 'message-id-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/ping',
                "comment": "Hi",
            })
            ok, pong = await transport.switch(ping)
            assert ok
            assert pong['@id'] == ping['@id']
    finally:
        await transport.stop()


@pytest.mark.asyncio
async def test_wallet_access(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        a2b = await get_pairwise(agent1, agent2)
        print('\n>START')
        stamp1 = datetime.now()
        for n in range(TEST_ITERATIONS):
            pw = await agent1.wallet.pairwise.get_pairwise(a2b.their.did)
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_decode_message(agent1: Agent):
    await agent1.open()
    try:
        seed = '000000000000000000000000000SEED1'
        packed = b'{"protected": "eyJlbmMiOiAieGNoYWNoYTIwcG9seTEzMDVfaWV0ZiIsICJ0eXAiOiAiSldNLzEuMCIsICJhbGciOiAiQXV0aGNyeXB0IiwgInJlY2lwaWVudHMiOiBbeyJlbmNyeXB0ZWRfa2V5IjogInBKcW1xQS1IVWR6WTNWcFFTb2dySGx4WTgyRnc3Tl84YTFCSmtHU2VMT014VUlwT0RQWTZsMVVsaVVvOXFwS0giLCAiaGVhZGVyIjogeyJraWQiOiAiM1ZxZ2ZUcDZRNFZlRjhLWTdlVHVXRFZBWmFmRDJrVmNpb0R2NzZLR0xtZ0QiLCAic2VuZGVyIjogIjRlYzhBeFRHcWtxamd5NHlVdDF2a0poeWlYZlNUUHo1bTRKQjk1cGZSMG1JVW9KajAwWmswNmUyUEVDdUxJYmRDck8xeTM5LUhGTG5NdW5YQVJZWk5rZ2pyYV8wYTBQODJpbVdNcWNHc1FqaFd0QUhOcUw1OGNkUUYwYz0iLCAiaXYiOiAiVU1PM2o1ZHZwQnFMb2Rvd3V0c244WEMzTkVqSWJLb2oifX1dfQ==", "iv": "MchkHF2M-4hneeUJ", "ciphertext": "UgcdsV-0rIkP25eJuRSROOuqiTEXp4NToKjPMmqqtJs-Ih1b5t3EEbrrHxeSfPsHtlO6J4OqA1jc5uuD3aNssUyLug==", "tag": "sQD8qgJoTrRoyQKPeCSBlQ=="}'
        await agent1.wallet.did.create_and_store_my_did(seed=seed)
        print('\n>START')
        stamp1 = datetime.now()
        for n in range(TEST_ITERATIONS):
            unpacked = await agent1.wallet.crypto.unpack_message(packed)
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()


@pytest.mark.asyncio
async def test_send_message(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        a2b = await get_pairwise(agent1, agent2)
        b2a = await get_pairwise(agent2, agent1)
        listener = await agent2.subscribe()

        print('\n>START')
        stamp1 = datetime.now()
        for n in range(TEST_ITERATIONS):
            msg = Message({
                '@id': 'message-id-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/message',
                "comment": "Hi. Are you listening?",
                "response_requested": True
            })
            await agent1.send_to(msg, a2b)
            resp = await listener.get_one()
            assert resp.message['@id'] == msg['@id']
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_send_message_via_transport(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        a2b = await get_pairwise(agent1, agent2)
        b2a = await get_pairwise(agent2, agent1)
        thread_id = 'thread-' + uuid.uuid4().hex
        transport_for_a = await agent1.spawn(thread_id, a2b)
        await transport_for_a.start()
        transport_for_b = await agent2.spawn(thread_id, b2a)
        await transport_for_b.start()

        print('\n>START')
        stamp1 = datetime.now()
        for n in range(TEST_ITERATIONS):
            msg = Message({
                '@id': 'message-id-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/message',
                "comment": "Hi. Are you listening?",
                "response_requested": True
            })
            await transport_for_a.send(msg)
            message, sender_vk, recip_vk = await transport_for_b.get_one()
            assert message['@id'] == msg['@id']
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_switch_via_transport_in_coros(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        a2b = await get_pairwise(agent1, agent2)
        b2a = await get_pairwise(agent2, agent1)
        thread_id = 'thread-' + uuid.uuid4().hex
        transport_for_a = await agent1.spawn(thread_id, a2b)
        await transport_for_a.start()
        transport_for_b = await agent2.spawn(thread_id, b2a)
        await transport_for_b.start()

        async def __producer(t: AbstractCoProtocolTransport):
            for n in range(TEST_ITERATIONS // 2):
                msg = Message({
                    '@id': 'message-id-' + uuid.uuid4().hex,
                    '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/message',
                    "comment": "RUN",
                    "response_requested": True
                })
                ok, resp = await t.switch(msg)
                assert ok is True
            msg = Message({
                '@id': 'message-id-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/message',
                "comment": "STOP",
                "response_requested": True
            })
            await t.send(msg)

        async def __consumer(t: AbstractCoProtocolTransport):
            message, sender_vk, recip_vk = await t.get_one()
            while True:
                ok, message = await t.switch(message)
                if message['comment'] == 'STOP':
                    return

        print('\n>START')
        stamp1 = datetime.now()
        producer = __producer(transport_for_a)
        consumer = __consumer(transport_for_b)
        await run_coroutines(producer, consumer, timeout=60)
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_send_message_via_transport_via_websocket(agent1: Agent, agent2: Agent):
    await agent1.open()
    await agent2.open()
    try:
        a2b = await get_pairwise(agent1, agent2)
        b2a = await get_pairwise(agent2, agent1)
        thread_id = 'thread-' + uuid.uuid4().hex
        a2b.their.endpoint = a2b.their.endpoint.replace('http://', 'ws://')
        transport_for_a = await agent1.spawn(thread_id, a2b)
        await transport_for_a.start()
        transport_for_b = await agent2.spawn(thread_id, b2a)
        await transport_for_b.start()

        print('\n>START')
        stamp1 = datetime.now()
        for n in range(TEST_ITERATIONS):
            msg = Message({
                '@id': 'message-id-' + uuid.uuid4().hex,
                '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test/1.0/message',
                "comment": "Hi. Are you listening?",
                "response_requested": True
            })
            await transport_for_a.send(msg)
            message, sender_vk, recip_vk = await transport_for_b.get_one()
            assert message['@id'] == msg['@id']
        print('\n>STOP')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'>timeout: {delta.seconds}')
    finally:
        await agent1.close()
        await agent2.close()



from typing import Optional, Any, List

from sirius_sdk.agent.wallet import NYMRole
from sirius_sdk.agent.wallet.abstract.ledger import AbstractLedger
from sirius_sdk.agent.connections import AgentRPC


class LedgerProxy(AbstractLedger):

    def __init__(self, rpc: AgentRPC):
        self.__rpc = rpc

    async def read_nym(self, pool_name: str, submitter_did: Optional[str], target_did: str) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/read_nym',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, target_did=target_did)
        )

    async def read_attribute(
            self, pool_name: str, submitter_did: Optional[str], target_did: str, name: str
    ) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/read_attribute',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, target_did=target_did, name=name)
        )

    async def write_nym(
            self, pool_name: str, submitter_did: str, target_did: str,
            ver_key: str = None, alias: str = None, role: NYMRole = None
    ) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/write_nym',
            params=dict(
                pool_name=pool_name, submitter_did=submitter_did,
                target_did=target_did, ver_key=ver_key, alias=alias, role=role
            )
        )

    async def register_schema(self, pool_name: str, submitter_did: str, data: dict) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/register_schema',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, data=data)
        )

    async def register_cred_def(self, pool_name: str, submitter_did: str, data: dict) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/register_cred_def',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, data=data)
        )

    async def write_attribute(
            self, pool_name: str, submitter_did: Optional[str], target_did: str, name: str, value: Any
    ) -> (bool, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/write_attribute',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, target_did=target_did, name=name, value=value)
        )

    async def sign_and_submit_request(self, pool_name: str, submitter_did: str, request: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/sign_and_submit_request',
            params=dict(pool_name=pool_name, submitter_did=submitter_did, request=request)
        )

    async def submit_request(self, pool_name: str, request: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/submit_request',
            params=dict(pool_name=pool_name, request=request)
        )

    async def submit_action(self, pool_name: str, request: dict, nodes: List[str] = None, timeout: int = None) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/submit_action',
            params=dict(pool_name=pool_name, request=request, nodes=nodes, timeout=timeout)
        )

    async def sign_request(self, submitter_did: str, request: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/sign_request',
            params=dict(submitter_did=submitter_did, request=request)
        )

    async def multi_sign_request(self, submitter_did: str, request: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/multi_sign_request',
            params=dict(submitter_did=submitter_did, request=request)
        )

    async def build_get_ddo_request(self, submitter_did: Optional[str], target_did: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_ddo_request',
            params=dict(submitter_did=submitter_did, target_did=target_did)
        )

    async def build_nym_request(
            self, submitter_did: str, target_did: str, ver_key: str = None, alias: str = None, role: NYMRole = None
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_nym_request',
            params=dict(
                submitter_did=submitter_did, target_did=target_did,
                ver_key=ver_key, alias=alias, role=role
            )
        )

    async def build_attrib_request(
            self, submitter_did: str, target_did: str, xhash: str = None, raw: dict = None, enc: str = None
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_attrib_request',
            params=dict(
                submitter_did=submitter_did,
                target_did=target_did, xhash=xhash, raw=raw, enc=enc
            )
        )

    async def build_get_attrib_request(
            self, submitter_did: Optional[str], target_did: str, raw: str = None, xhash: str = None, enc: str = None
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_attrib_request',
            params=dict(submitter_did=submitter_did, target_did=target_did, raw=raw, xhash=xhash, enc=enc)
        )

    async def build_get_nym_request(self, submitter_did: Optional[str], target_did: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_nym_request',
            params=dict(submitter_did=submitter_did, target_did=target_did)
        )

    async def parse_get_nym_response(self, response: Any) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_nym_response',
            params=dict(response=response)
        )

    async def build_schema_request(self, submitter_did: str, data: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_schema_request',
            params=dict(submitter_did=submitter_did, data=data)
        )

    async def build_get_schema_request(self, submitter_did: Optional[str], id_: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_schema_request',
            params=dict(submitter_did=submitter_did, id_=id_)
        )

    async def parse_get_schema_response(self, get_schema_response: dict) -> (str, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_schema_response',
            params=dict(get_schema_response=get_schema_response)
        )

    async def build_cred_def_request(self, submitter_did: str, data: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/admin/1.0/build_cred_def_request',
            params=dict(submitter_did=submitter_did, data=data)
        )

    async def build_get_cred_def_request(self, submitter_did: Optional[str], id_: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_cred_def_request',
            params=dict(submitter_did=submitter_did, id_=id_)
        )

    async def parse_get_cred_def_response(self, get_cred_def_response: dict) -> (str, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_cred_def_response',
            params=dict(get_cred_def_response=get_cred_def_response)
        )

    async def build_node_request(self, submitter_did: str, target_did: str, data: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_node_request',
            params=dict(submitter_did=submitter_did, target_did=target_did, data=data)
        )

    async def build_get_validator_info_request(self, submitter_did: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_validator_info_request',
            params=dict(submitter_did=submitter_did)
        )

    async def build_get_txn_request(
            self, submitter_did: Optional[str], ledger_type: Optional[str], seq_no: int
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_txn_request',
            params=dict(submitter_did=submitter_did, ledger_type=ledger_type, seq_no=seq_no)
        )

    async def build_pool_config_request(self, submitter_did: str, writes: bool, force: bool) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_pool_config_request',
            params=dict(submitter_did=submitter_did, writes=writes, force=force)
        )

    async def build_pool_restart_request(self, submitter_did: str, action: str, datetime: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_pool_restart_request',
            params=dict(submitter_did=submitter_did, action=action, datetime=datetime)
        )

    async def build_pool_upgrade_request(
            self, submitter_did: str, name: str, version: str, action: str, _sha256: str,
            _timeout: Optional[int], schedule: Optional[str], justification: Optional[str],
            reinstall: bool, force: bool, package: Optional[str]
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_pool_upgrade_request',
            params=dict(
                submitter_did=submitter_did, name=name, version=version, action=action,
                _sha256=_sha256, _timeout=_timeout, schedule=schedule, justification=justification,
                reinstall=reinstall, force=force, package=package
            )
        )

    async def build_revoc_reg_def_request(self, submitter_did: str, data: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_revoc_reg_def_request',
            params=dict(submitter_did=submitter_did, data=data)
        )

    async def build_get_revoc_reg_def_request(
            self, submitter_did: Optional[str], rev_reg_def_id: str
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_revoc_reg_def_request',
            params=dict(submitter_did=submitter_did, rev_reg_def_id=rev_reg_def_id)
        )

    async def parse_get_revoc_reg_def_response(self, get_revoc_ref_def_response: dict) -> (str, dict):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_revoc_reg_def_response',
            params=dict(get_revoc_ref_def_response=get_revoc_ref_def_response)
        )

    async def build_revoc_reg_entry_request(
            self, submitter_did: str, revoc_reg_def_id: str, rev_def_type: str, value: dict
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_revoc_reg_entry_request',
            params=dict(
                submitter_did=submitter_did, revoc_reg_def_id=revoc_reg_def_id, rev_def_type=rev_def_type, value=value
            )
        )

    async def build_get_revoc_reg_request(
            self, submitter_did: Optional[str], revoc_reg_def_id: str, timestamp: int
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_revoc_reg_request',
            params=dict(submitter_did=submitter_did, revoc_reg_def_id=revoc_reg_def_id, timestamp=timestamp)
        )

    async def parse_get_revoc_reg_response(self, get_revoc_reg_response: dict) -> (str, dict, int):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_revoc_reg_response',
            params=dict(get_revoc_reg_response=get_revoc_reg_response)
        )

    async def build_get_revoc_reg_delta_request(
            self, submitter_did: Optional[str], revoc_reg_def_id: str,
            from_: Optional[int], to: int
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_revoc_reg_delta_request',
            params=dict(
                submitter_did=submitter_did, revoc_reg_def_id=revoc_reg_def_id, from_=from_, to=to
            )
        )

    async def parse_get_revoc_reg_delta_response(self, get_revoc_reg_delta_response: dict) -> (str, dict, int):
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/parse_get_revoc_reg_delta_response',
            params=dict(get_revoc_reg_delta_response=get_revoc_reg_delta_response)
        )

    async def get_response_metadata(self, response: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/get_response_metadata',
            params=dict(response=response)
        )

    async def build_auth_rule_request(
            self, submitter_did: str, txn_type: str, action: str, field: str,
            old_value: Optional[str], new_value: Optional[str], constraint: dict
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_auth_rule_request',
            params=dict(
                submitter_did=submitter_did, txn_type=txn_type, action=action, field=field,
                old_value=old_value, new_value=new_value, constraint=constraint
            )
        )

    async def build_auth_rules_request(self, submitter_did: str, data: dict) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_auth_rules_request',
            params=dict(submitter_did=submitter_did, data=data)
        )

    async def build_get_auth_rule_request(
            self, submitter_did: Optional[str], txn_type: Optional[str],
            action: Optional[str], field: Optional[str], old_value: Optional[str], new_value: Optional[str]
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_auth_rule_request',
            params=dict(
                submitter_did=submitter_did, txn_type=txn_type, action=action,
                field=field, old_value=old_value, new_value=new_value
            )
        )

    async def build_txn_author_agreement_request(
            self, submitter_did: str, text: Optional[str],
            version: str, ratification_ts: Optional[int] = None, retirement_ts: Optional[int] = None
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_txn_author_agreement_request',
            params=dict(
                submitter_did=submitter_did, text=text, version=version,
                ratification_ts=ratification_ts, retirement_ts=retirement_ts
            )
        )

    async def build_disable_all_txn_author_agreements_request(self, submitter_did: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_disable_all_txn_author_agreements_request',
            params=dict(submitter_did=submitter_did)
        )

    async def build_get_txn_author_agreement_request(self, submitter_did: Optional[str], data: dict = None) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_txn_author_agreement_request',
            params=dict(submitter_did=submitter_did, data=data)
        )

    async def build_acceptance_mechanisms_request(
            self, submitter_did: str, aml: dict, version: str, aml_context: Optional[str]
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_acceptance_mechanisms_request',
            params=dict(submitter_did=submitter_did, aml=aml, version=version, aml_context=aml_context)
        )

    async def build_get_acceptance_mechanisms_request(
            self, submitter_did: Optional[str], timestamp: Optional[int], version: Optional[str]
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/build_get_acceptance_mechanisms_request',
            params=dict(submitter_did=submitter_did, timestamp=timestamp, version=version)
        )

    async def append_txn_author_agreement_acceptance_to_request(
            self, request: dict, text: Optional[str],
            version: Optional[str], taa_digest: Optional[str], mechanism: str, time: int
    ) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/append_txn_author_agreement_acceptance_to_request',
            params=dict(
                request=request, text=text, version=version,
                taa_digest=taa_digest, mechanism=mechanism, time=time
            )
        )

    async def append_request_endorser(self, request: dict, endorser_did: str) -> dict:
        return await self.__rpc.remote_call(
            msg_type='did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/sirius_rpc/1.0/append_request_endorser',
            params=dict(request=request, endorser_did=endorser_did)
        )



from sirius_sdk.agent.agent import Agent
from sirius_sdk.hub import init, context, endpoints, ledger, subscribe, ping, send, send_to, \
    generate_qr_code, DID, Crypto, Microledgers, PairwiseList, AnonCreds, CoProtocolThreadedP2P, CoProtocolP2PAnon, \
    CoProtocolP2P, AbstractP2PCoProtocol, CoProtocolThreadedTheirs, Cache, open_communication, NonSecrets, \
    acquire, release
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.agent.pairwise import Pairwise, TheirEndpoint
from sirius_sdk.agent.connections import Endpoint
from sirius_sdk.agent import aries_rfc
from sirius_sdk.agent.ledger import Schema, CredentialDefinition, AnonCredSchema, Ledger
from sirius_sdk.errors import indy_exceptions, exceptions


__all__ = [
    "Agent", "P2PConnection", "Pairwise", "TheirEndpoint", "Endpoint", "init", "context", "endpoints", "ledger",
    "subscribe", "ping", "send", "send_to", "generate_qr_code", "DID", "Crypto", "Microledgers", "PairwiseList",
    "CoProtocolThreadedP2P", "CoProtocolP2PAnon", "CoProtocolP2P", "AbstractP2PCoProtocol",
    "aries_rfc", "CoProtocolThreadedTheirs", "AnonCreds", "Cache", "open_communication", "NonSecrets",
    "Schema", "CredentialDefinition", "AnonCredSchema", "indy_exceptions", "exceptions", "Ledger", "acquire", "release"
]



import sys
from abc import ABC, abstractmethod
from typing import List, Optional
from urllib.parse import urlparse, urlunparse

from sirius_sdk.agent.wallet.abstract.did import AbstractDID
from sirius_sdk.agent.wallet.abstract.pairwise import AbstractPairwise


class TheirEndpoint:

    def __init__(self, endpoint: str, verkey: str, routing_keys: List[str]=None):
        self.endpoint = endpoint
        self.verkey = verkey
        self.routing_keys = routing_keys or []

    @property
    def netloc(self) -> Optional[str]:
        if self.endpoint:
            return urlparse(self.endpoint).netloc
        else:
            return None

    @netloc.setter
    def netloc(self, value: str):
        if self.endpoint:
            components = list(urlparse(self.endpoint))
            components[1] = value
            self.endpoint = urlunparse(components)


class Pairwise:

    class Their(TheirEndpoint):

        def __init__(
                self, did: str, label: str, endpoint: str, verkey: str,
                routing_keys: List[str] = None, did_doc: dict = None
        ):
            self.did = did
            self.label = label
            self.did_doc = did_doc
            super().__init__(endpoint, verkey, routing_keys)

    class Me:

        def __init__(self, did, verkey, did_doc: dict = None):
            self.did = did
            self.verkey = verkey
            self.did_doc = did_doc

        def __eq__(self, other):
            if isinstance(other, Pairwise.Me):
                return self.did == other.did and self.verkey == other.verkey and self.did_doc == other.did_doc

    def __init__(self, me: Me, their: Their, metadata: dict=None):
        self.__me = me
        self.__their = their
        self.__metadata = metadata

    @property
    def their(self) -> Their:
        return self.__their

    @property
    def me(self) -> Me:
        return self.__me

    @property
    def metadata(self) -> dict:
        return self.__metadata


class AbstractPairwiseList(ABC):

    @abstractmethod
    async def create(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def update(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def is_exists(self, their_did: str) -> bool:
        raise NotImplemented

    @abstractmethod
    async def ensure_exists(self, pairwise: Pairwise):
        raise NotImplemented

    @abstractmethod
    async def load_for_did(self, their_did: str) -> Optional[Pairwise]:
        raise NotImplemented

    @abstractmethod
    async def load_for_verkey(self, their_verkey: str) -> Optional[Pairwise]:
        raise NotImplemented

    async def enumerate(self):
        cur = 0
        await self._start_loading()
        try:
            while True:
                success, collection = await self._partial_load()
                if success:
                    for p in collection:
                        yield cur, p
                        cur += 1
                else:
                    break
        finally:
            await self._stop_loading()

    @abstractmethod
    async def _start_loading(self):
        raise NotImplemented

    @abstractmethod
    async def _partial_load(self) -> (bool, List[Pairwise]):
        raise NotImplemented

    @abstractmethod
    async def _stop_loading(self):
        raise NotImplemented

    def __aiter__(self):
        return self


class WalletPairwiseList(AbstractPairwiseList):

    def __init__(self, api: (AbstractPairwise, AbstractDID)):
        self._api_pairwise = api[0]
        self._api_did = api[1]
        self.__is_loading = False

    async def create(self, pairwise: Pairwise):
        await self._api_did.store_their_did(did=pairwise.their.did, verkey=pairwise.their.verkey)
        metadata = pairwise.metadata or {}
        metadata.update(self._build_metadata(pairwise))
        await self._api_pairwise.create_pairwise(
            their_did=pairwise.their.did,
            my_did=pairwise.me.did,
            metadata=metadata,
            tags=self._build_tags(pairwise)
        )

    async def update(self, pairwise: Pairwise):
        await self._api_pairwise.set_pairwise_metadata(
            their_did=pairwise.their.did,
            metadata=pairwise.metadata,
            tags=self._build_tags(pairwise)
        )

    async def is_exists(self, their_did: str) -> bool:
        return await self._api_pairwise.is_pairwise_exists(their_did=their_did)

    async def ensure_exists(self, pairwise: Pairwise):
        if await self.is_exists(their_did=pairwise.their.did):
            await self.update(pairwise)
        else:
            await self.create(pairwise)

    async def load_for_did(self, their_did: str) -> Optional[Pairwise]:
        if await self.is_exists(their_did):
            raw = await self._api_pairwise.get_pairwise(their_did)
            metadata = raw['metadata']
            pairwise = self._restore_pairwise(metadata)
            return pairwise
        else:
            return None

    async def load_for_verkey(self, their_verkey: str) -> Optional[Pairwise]:
        collection, count = await self._api_pairwise.search(tags={'their_verkey': their_verkey}, limit=1)
        if collection:
            metadata = collection[0]['metadata']
            pairwise = self._restore_pairwise(metadata)
            return pairwise
        else:
            return None

    async def _start_loading(self):
        self.__is_loading = True

    async def _partial_load(self) -> (bool, List[Pairwise]):
        if self.__is_loading:
            items = await self._api_pairwise.list_pairwise()
            self.__is_loading = False
            return True, [self._restore_pairwise(item['metadata']) for item in items]
        else:
            return False, []

    async def _stop_loading(self):
        self.__is_loading = False

    @staticmethod
    def _build_tags(p: Pairwise):
        return {
            'my_did': p.me.did,
            'my_verkey': p.me.verkey,
            'their_verkey': p.their.verkey
        }

    @staticmethod
    def _restore_pairwise(metadata: dict):
        pairwise = Pairwise(
            me=Pairwise.Me(
                did=metadata.get('me', {}).get('did', None),
                verkey=metadata.get('me', {}).get('verkey', None),
                did_doc=metadata.get('me', {}).get('did_doc', None)
            ),
            their=Pairwise.Their(
                did=metadata.get('their', {}).get('did', None),
                verkey=metadata.get('their', {}).get('verkey', None),
                label=metadata.get('their', {}).get('label', None),
                endpoint=metadata.get('their', {}).get('endpoint', {}).get('address', None),
                routing_keys=metadata.get('their', {}).get('endpoint', {}).get('routing_keys', None),
                did_doc=metadata.get('their', {}).get('did_doc', None)
            ),
            metadata=metadata
        )
        return pairwise

    @staticmethod
    def _build_metadata(pairwise: Pairwise) -> dict:
        metadata = {
            'me': {
                'did': pairwise.me.did,
                'verkey': pairwise.me.verkey,
                'did_doc': pairwise.me.did_doc
            },
            'their': {
                'did': pairwise.their.did,
                'verkey': pairwise.their.verkey,
                'label': pairwise.their.label,
                'endpoint': {
                    'address': pairwise.their.endpoint,
                    'routing_keys': pairwise.their.routing_keys
                },
                'did_doc': pairwise.their.did_doc
            }
        }
        return metadata



from sirius_sdk.messaging import *
from sirius_sdk.agent.aries_rfc.feature_0048_trust_ping.messages import Ping, Pong
from sirius_sdk.agent.aries_rfc.feature_0015_acks.messages import Ack, Status as AckStatus


class Test1Message(Message):
    pass


class Test2Message(Message):
    pass


def test_type_parsing():
    str1 = 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test-protocol/1.0/name'
    typ = Type.from_str(str1)
    assert typ.doc_uri == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/'
    assert typ.protocol == 'test-protocol'
    assert typ.name == 'name'
    assert typ.version == '1.0'
    assert typ.version_info.major == 1 and typ.version_info.minor == 0 and typ.version_info.patch == 0

    str2 = 'https://didcomm.org/test-protocol/1.2/name'
    typ = Type.from_str(str2)
    assert typ.doc_uri == 'https://didcomm.org/'
    assert typ.protocol == 'test-protocol'
    assert typ.name == 'name'
    assert typ.version == '1.2'
    assert typ.version_info.major == 1 and typ.version_info.minor == 2 and typ.version_info.patch == 0


def test_register_protocol_message_success():
    register_message_class(Test1Message, protocol='test-protocol')
    ok, msg = restore_message_instance(
        {
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test-protocol/1.0/name'
        }
    )
    assert ok is True
    assert isinstance(msg, Test1Message)


def test_agnostic_doc_uri():
    register_message_class(Test1Message, protocol='test-protocol')
    ok, msg = restore_message_instance(
        {
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test-protocol/1.0/name'
        }
    )
    assert ok is True
    assert isinstance(msg, Test1Message)

    ok, msg = restore_message_instance(
        {
            '@type': 'https://didcomm.org/test-protocol/1.0/name'
        }
    )
    assert ok is True
    assert isinstance(msg, Test1Message)


def test_register_protocol_message_fail():
    register_message_class(Test1Message, protocol='test-protocol')
    ok, msg = restore_message_instance(
        {
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/fake-protocol/1.0/name'
        }
    )
    assert ok is False
    assert msg is None


def test_register_protocol_message_multiple_name():
    register_message_class(Test1Message, protocol='test-protocol')
    register_message_class(Test2Message, protocol='test-protocol', name='test-name')
    ok, msg = restore_message_instance(
        {
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test-protocol/1.0/name'
        }
    )
    assert ok is True
    assert isinstance(msg, Test1Message)

    ok, msg = restore_message_instance(
        {
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/test-protocol/1.0/test-name'
        }
    )
    assert ok is True
    assert isinstance(msg, Test2Message)


def test_aries_ping_pong():
    ok, ping = restore_message_instance(
        {
            '@id': 'trust-ping-message-id',
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
            "comment": "Hi. Are you OK?",
            "response_requested": True
        }
    )
    assert ok is True
    assert isinstance(ping, Ping)
    assert ping.comment == 'Hi. Are you OK?'
    assert ping.response_requested is True

    ok, pong = restore_message_instance(
        {
            '@id': 'trust-ping_response-message-id',
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping_response',
            "comment": "Hi. I am OK!",
            "~thread": {
                "thid": "ping-id"
            }
        }
    )
    assert ok is True
    assert isinstance(pong, Pong)
    assert pong.comment == 'Hi. I am OK!'
    assert pong.ping_id == 'ping-id'


def test_aries_ack():

    message = Ack(thread_id='ack-thread-id', status=AckStatus.PENDING)
    assert message.protocol == 'notification'
    assert message.name == 'ack'
    assert message.version == '1.0'
    assert str(message.version_info) == '1.0.0'
    assert message.status == AckStatus.PENDING
    message.validate()

    ok, ack = restore_message_instance(
        {
            '@id': 'ack-message-id',
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/notification/1.0/ack',
            'status': 'PENDING',
            "~thread": {
                'thid': 'thread-id'
            },
        }
    )
    assert ok is True
    assert isinstance(ack, Ack)
    assert ack.thread_id == 'thread-id'
    ack.validate()
    assert ack.status == AckStatus.PENDING

